"""
Script to retrieve the data from tensorboard logs.
It computes the maximum or minimum per each time series and saves it in a file.
It is intended to be used to process the gnn results.


Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
import os
import numpy as np
import pandas as pd
from typing import Dict
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator


def extract_scalar_from_tb(path: str) -> Dict[str, float]:
    """
    Extracts the best values of all the time series.

    :param path: Path to tensorboard file.
    :type path: str

    :return: Dictionary with metrics as keys and maximums as values.
    :rtype: Dict[str, float]
    """
    event_acc = EventAccumulator(path)
    event_acc.Reload()
    res = {}
    for scalar_key in event_acc.scalars.Keys():
        scalar_values = list(map(lambda x: x.value, event_acc.Scalars(scalar_key)))
        if 'ECE' in scalar_key or 'Perc' in scalar_key:
            scalar_best = np.min(scalar_values)
        else:
            scalar_best = np.max(scalar_values)
        assert scalar_key not in res, 'More than one line per metric is not supported.'
        res[scalar_key] = scalar_best
    return res


def extract_all_scalars(path: str) -> Dict[str, Dict[str, float]]:
    """
    Extracts all the best values of all metrics and files.
    Tensorboard files must be under a folder.
    Folders' names will be used as dictionary keys.

    :param path: Path to folder containing folders containing tensorboard files.
    :type path: str

    :return: Dictionary with folders' names as keys, and another dictionary as values. That dictionary contains metrics and keys and maximums as values.
    :rtype: Dict[str, Dict[str, float]]
    """
    res = {}
    for tb_input in filter(
            lambda x: os.path.isdir(os.path.join(path, x)),
            os.listdir(path)
            ):
        tb_files = os.listdir(os.path.join(path, tb_input))
        assert len(tb_files) == 1, 'There must be only one tensorboard file per folder.'
        tb_file = os.path.join(path, tb_input, tb_files[0])
        scalar_values = extract_scalar_from_tb(tb_file)
        res[tb_input] = scalar_values
    return res


def to_long_format(values: pd.DataFrame) -> pd.DataFrame:
    """
    Converts the initial dataframe into a more useful one that is in long format.
    New columns: arch, bn, dropout, layers, metric, value

    :param values: The initial dataframe from extract_tensorboard.
    :type values: pd.DataFrame

    :return: DataFrame in long format.
    :rtype: pd.DataFrame
    """
    data = values.copy()
    data = data.reset_index()
    data[['arch', 'layers', 'dropout', 'bn']] = data['index'].str.split('_', expand=True)
    data['layers'] = pd.to_numeric(data['layers'])
    data['dropout'] = pd.to_numeric(data['dropout'])
    data['bn'] = data['bn'].apply(lambda x: {'bn': 'Yes', 'None': 'No'}.get(x, 'Unknown'))
    data = data.drop(columns=['index'])
    data = data.melt(id_vars=['arch', 'layers', 'dropout', 'bn'], var_name='metric')
    data[['metric', 'split']] = data['metric'].str.split('/', expand=True)
    return data


def write_to_file(path: str, values: Dict[str, Dict[str, float]]) -> None:
    """
    Saves output to file. Dictionary keys are used as index.
    Inner dictionary keys are used as columns.

    :param path: Path to save file, without extension, .csv appended.
    :type path: str
    :param values: Values to save.
    :type values: Dict[str, Dict[str, float]]
    """
    values_df = pd.DataFrame(values).transpose()
    values_df = to_long_format(values_df)
    values_df.to_csv(path + '.csv', index=False)


def _create_parser():
    parser = argparse.ArgumentParser(description='Process tensorboard gnn logs.')
    parser.add_argument('--logs-dir', required=True, help='Path to gnn logs dir with tensorboard files.')
    parser.add_argument('--output-path', required=True, help='Output file path, must not have extension.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()

    metrics = extract_all_scalars(args.logs_dir)
    write_to_file(args.output_path, metrics)



"""
Draws the cells into an image.
Input format: PNG / CSV
Output format: PNG

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
from tqdm import tqdm
import os
import cv2
import numpy as np
import pandas as pd
import json
from typing import Dict, Tuple, List
from ..utils.preprocessing import get_names, read_labels


def draw_cells(
        orig: np.ndarray,
        png: np.ndarray,
        csv: pd.DataFrame,
        type_info: Dict[str, Tuple[str, List[int]]]
        ) -> np.ndarray:
    """
    Draws cell labels on the original image based on the PNG labels and CSV data.

    Given the original image, PNG labels, CSV data containing cell labels,
    and type information, this function draws the cell labels on the original image.
    The cell labels are blended with the original image by alpha compositing.

    :param orig: The original image.
    :type orig: np.ndarray
    :param png: The PNG labels representing cell IDs.
    :type png: np.ndarray
    :param csv: The CSV data containing cell labels.
    :type csv: pd.DataFrame
    :param type_info: The type information dictionary mapping cell labels to colors.
    :type type_info: Dict[str, Tuple[str, List[int]]]
    :return: The image with cell labels drawn on it.
    :rtype: np.ndarray
    """
    blend = orig.copy()
    for i, (idx, cell_label) in csv.iterrows():
        blend[png == idx] = 0.3 * np.array(type_info[str(cell_label)][1]) \
            + 0.7 * blend[png == idx]
    return blend


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--orig-dir', type=str, help='Path to base images folder. Must be in .png format.')
    parser.add_argument('--png-dir', type=str, help='Path to folder with png of the labels.')
    parser.add_argument('--csv-dir', type=str, help='Path to folder with csv of the labels.')
    parser.add_argument('--output-dir', type=str, help='Path to folder where to save results.')
    parser.add_argument('--type-info', type=str, help='Path to type_info.json.')
    return parser


def main_with_args(args: Namespace) -> None:
    os.makedirs(args.output_dir, exist_ok=True)
    names = get_names(args.orig_dir, '.png')
    with open(args.type_info, 'r') as f:
            type_info = json.load(f)
    for name in tqdm(names):
        orig = cv2.imread(os.path.join(args.orig_dir, name + '.png'), cv2.IMREAD_COLOR)[:, :, ::-1]
        png, csv = read_labels(name, args.png_dir, args.csv_dir)
        out = draw_cells(orig, png, csv, type_info)
        cv2.imwrite(os.path.join(args.output_dir, name + '.overlay.png'), out[:, :, ::-1])
    return


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Preprocess data into hovernet input format.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import cv2
import numpy as np
import pandas as pd
import random
from typing import Optional, List, Tuple, Set, Callable
import argparse
from tqdm import tqdm
import os
from tumourkit.utils.preprocessing import parse_path, create_dir, get_names


def to4dim(orig_dir: str, png_dir: str, csv_dir: str, name: str) -> np.ndarray:
    """
    Generates 3D numpy array where dimensions are:
    Height x Width x Channels
    Channels are: R, G, B, Ground Truth (segmentation)
    csv_dir is only needed for compatibility, it is not used.
    """
    gt_path = png_dir + name + '.GT_cells.png'
    img_path = orig_dir + name + '.png'

    gt = cv2.imread(gt_path, -1)
    gt = np.reshape(gt, (*gt.shape, 1))
    img = cv2.imread(img_path, -1)[:,:,::-1]
    return np.concatenate([img, gt], axis=-1)


def generate_labels(
        png_dir: str, csv_dir: str, name: str,
        return_segm: Optional[bool]=True) -> np.ndarray:
    """
    Generates class GT. Pixels have values:
    0-> Background
    1-> Non-tumour
    2-> Tumour
    Return shape: Height x Width x 1
    """
    gt_path = png_dir + name + '.GT_cells.png'
    gt = cv2.imread(gt_path, -1)
    gt = np.reshape(gt, (*gt.shape, 1))
    
    gt_labels_path = csv_dir + name + '.class.csv'
    parser = pd.read_csv(gt_labels_path, header=None)
    parser = dict(zip(parser.iloc[:,0], parser.iloc[:,1]))

    def aux(x):
        if x>1:
            return parser[x]
        return x
    vec_parser = np.vectorize(aux)
    
    if return_segm:
        return gt, vec_parser(gt)
    return vec_parser(gt)


def to5dim(orig_dir: str, png_dir: str, csv_dir: str, name: str) -> np.ndarray:
    """
    Generates 3D numpy array where dimensions are:
    Height x Width x Channels
    Channels are: R, G, B, GT (segmentation), GT (class)
    """
    gt, labels = generate_labels(png_dir, csv_dir, name)

    img_path = orig_dir + name + '.png'
    img = cv2.imread(img_path, -1)[:,:,::-1]

    return np.concatenate([img, gt, labels], axis=-1)


def create_splits(
        img_list: List[str], out_dir: str
        ) -> Tuple[Tuple[Set[str], Set[str], Set[str]], List[str]]:
    """
    img_list: List of image names to split.
    out_dir: Parent directory to create subfolders train, validation and test.

    Returns sets of strings as splits, and creates subfolders
    for saving data on them later.
    """
    split_dirs = [out_dir + x for x in ['train/', 'validation/', 'test/']]
    for dir_ in split_dirs:
        if not os.path.isdir(dir_):
            os.mkdir(dir_)
    n = len(img_list)
    n_train = int(n * 0.7)
    n_val = int(n * 0.25)
    n_test = n - n_train - n_val
    print('Train samples:', n_train)
    print('Validation samples:', n_val)
    print('Test samples:', n_test)

    aux = set(img_list)
    train_split = set(random.sample(aux, n_train))
    aux = aux.difference(train_split)
    val_split = set(random.sample(aux, n_val))
    test_split = aux.difference(val_split)

    splits = [train_split, val_split, test_split]
    return splits, split_dirs


def save_batch(
        name: str, save_dir: str,
        processing_function: Callable[[str, str, str, str], np.ndarray],
        orig_dir: str, png_dir: str, csv_dir: str,
        shape: str) -> None:
    """
    Loads data from paths and saves the joined array in npy format.
    Three output shapes are supported:
    'full': The whole image is saved.
    '518': Image is pieced into 518x518 patches.
    '270': Image is pieced into 270x270 patches.
    """
    arr = processing_function(orig_dir, png_dir, csv_dir, name)
    if shape == 'full':
        np.save(save_dir + name + '.npy', arr)
    elif shape == '270':
        for i in [0, 252, 503, 754]:
            for j in [0, 252, 503, 754]:
                crop = arr[i:i+270, j:j+270,]
                np.save(save_dir + name + '(' + str(i) + ',' + str(j) + ').npy', crop)
    elif shape == '518':
        for i in [0, 506]:
            for j in [0, 506]:
                crop = arr[i:i+518, j:j+518,]
                np.save(save_dir + name + '(' + str(i) + ',' + str(j) + ').npy', crop)
    else:
        assert False, 'Something wrong happened.'


def save_npy(
        img_list: str, out_dir: str,
        processing_function: Callable[[str, str, str, str], np.ndarray],
        split: bool, orig_dir: str, png_dir: str, csv_dir: str,
        shape: bool) -> None:
    """
    Saves data in the npy required format.
    Internally calls save_batch for each split if wanted.
    """
    if split:
        splits, split_dirs = create_splits(img_list, out_dir)
        for split_, split_dir in zip(splits, split_dirs):
            print('Split path: ', split_dir)
            for name in tqdm(split_):
                save_batch(name, split_dir, processing_function, orig_dir, png_dir, csv_dir, shape)

    else:
        for name in tqdm(img_list):
            save_batch(name, out_dir, processing_function, orig_dir, png_dir, csv_dir, shape)


def save_imgs(
        orig_dir: str, png_dir: str, csv_dir: str,
        name: str, out_dir: str) -> None:
    """
    Saves RGB image, GT segmentation image and GT class image
    in png format.
    """
    arr = to5dim(orig_dir, png_dir, csv_dir, name)
    img = np.array(arr[:,:,:3][:,:,::-1], dtype=np.uint8)
    gt = arr[:,:,3]
    labels = arr[:,:,4]

    save_dir = out_dir+'samples/'
    create_dir(save_dir)
    cv2.imwrite(save_dir + name + '.png', img)
    cv2.imwrite(save_dir + name + '.GT_cells.png', gt)
    cv2.imwrite(save_dir + name + '.class.png', labels)


def _create_parser():
    parser = argparse.ArgumentParser(description='Generates .npy files for HoVer-net from the current data images in .png and .csv.')
    parser.add_argument('--orig-dir', type=str, required=True, help='Directory to read the original png files. Must end with a slash /.')
    parser.add_argument('--png-dir', type=str, required=True, help='Directory to read the label png files. Must end with a slash /.')
    parser.add_argument('--csv-dir', type=str, required=True, help='Directory to read the label csv files. Must end with a slash /.')
    parser.add_argument('--out-dir', type=str, required=True, help='Directory to save the .npy files. Must end with a slash /.')
    parser.add_argument('--save-example', action='store_true', help="Save one sample as images to visualize.")
    parser.add_argument('--use-labels', action='store_true', help="Whether to generate the label image or only segmentation.")
    parser.add_argument('--split', action='store_true', help="Whether to split into train and validation folders.")
    parser.add_argument('--shape', type=str, choices=['270', '518', 'full'], required=True, help='Height and width of the saved arrays. If less than full, patching is applied.')
    return parser


def main_with_args(args):
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)
    orig_dir = parse_path(args.orig_dir)
    out_dir = parse_path(args.out_dir)
    create_dir(out_dir)

    img_list = get_names(png_dir, '.GT_cells.png')

    if args.save_example:
        name = random.sample(img_list, 1)[0]
        save_imgs(orig_dir, png_dir, csv_dir, name, out_dir)

    if args.use_labels:
        save_npy(img_list, out_dir, to5dim, args.split, orig_dir, png_dir, csv_dir, args.shape)
    else:
        save_npy(img_list, out_dir, to4dim, args.split, orig_dir, png_dir, csv_dir, args.shape)

def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)




import cv2
import numpy as np
from scipy.ndimage import measurements

####
def fix_mirror_padding(ann):
    """Deal with duplicated instances due to mirroring in interpolation
    during shape augmentation (scale, rotation etc.).
    
    """
    current_max_id = np.amax(ann)
    inst_list = list(np.unique(ann))
    inst_list.remove(0)  # 0 is background
    for inst_id in inst_list:
        inst_map = np.array(ann == inst_id, np.uint8)
        remapped_ids = measurements.label(inst_map)[0]
        remapped_ids[remapped_ids > 1] += current_max_id
        ann[remapped_ids > 1] = remapped_ids[remapped_ids > 1]
        current_max_id = np.amax(ann)
    return ann


####
def gaussian_blur(images, random_state, parents, hooks, max_ksize=3):
    """Apply Gaussian blur to input images."""
    img = images[0]  # aleju input batch as default (always=1 in our case)
    ksize = random_state.randint(0, max_ksize, size=(2,))
    ksize = tuple((ksize * 2 + 1).tolist())

    ret = cv2.GaussianBlur(
        img, ksize, sigmaX=0, sigmaY=0, borderType=cv2.BORDER_REPLICATE
    )
    ret = np.reshape(ret, img.shape)
    ret = ret.astype(np.uint8)
    return [ret]


####
def median_blur(images, random_state, parents, hooks, max_ksize=3):
    """Apply median blur to input images."""
    img = images[0]  # aleju input batch as default (always=1 in our case)
    ksize = random_state.randint(0, max_ksize)
    ksize = ksize * 2 + 1
    ret = cv2.medianBlur(img, ksize)
    ret = ret.astype(np.uint8)
    return [ret]


####
def add_to_hue(images, random_state, parents, hooks, range=None):
    """Perturbe the hue of input images."""
    img = images[0]  # aleju input batch as default (always=1 in our case)
    hue = random_state.uniform(*range)
    hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    if hsv.dtype.itemsize == 1:
        # OpenCV uses 0-179 for 8-bit images
        hsv[..., 0] = (hsv[..., 0] + hue) % 180
    else:
        # OpenCV uses 0-360 for floating point images
        hsv[..., 0] = (hsv[..., 0] + 2 * hue) % 360
    ret = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)
    ret = ret.astype(np.uint8)
    return [ret]


####
def add_to_saturation(images, random_state, parents, hooks, range=None):
    """Perturbe the saturation of input images."""
    img = images[0]  # aleju input batch as default (always=1 in our case)
    value = 1 + random_state.uniform(*range)
    gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    ret = img * value + (gray * (1 - value))[:, :, np.newaxis]
    ret = np.clip(ret, 0, 255)
    ret = ret.astype(np.uint8)
    return [ret]


####
def add_to_contrast(images, random_state, parents, hooks, range=None):
    """Perturbe the contrast of input images."""
    img = images[0]  # aleju input batch as default (always=1 in our case)
    value = random_state.uniform(*range)
    mean = np.mean(img, axis=(0, 1), keepdims=True)
    ret = img * value + mean * (1 - value)
    ret = np.clip(img, 0, 255)
    ret = ret.astype(np.uint8)
    return [ret]


####
def add_to_brightness(images, random_state, parents, hooks, range=None):
    """Perturbe the brightness of input images."""
    img = images[0]  # aleju input batch as default (always=1 in our case)
    value = random_state.uniform(*range)
    ret = np.clip(img + value, 0, 255)
    ret = ret.astype(np.uint8)
    return [ret]



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os

from tumourkit.utils.read_nodes import create_node_splits
from tumourkit.utils.preprocessing import parse_path 

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
GRAPHS_DIR = parse_path(TEST_DIR) + 'graphs/'


@pytest.mark.parametrize("val_size,test_size", [
    (0.2,0.2), (0.2,0.1), (0.8,0.1), (0.1,0.8), (0.45,0.45)
])
def test_graph_centroids(val_size, test_size):
    X_tr, X_val, X_ts, y_tr, y_val, y_ts = \
        create_node_splits(GRAPHS_DIR, val_size, test_size, mode='total')
    X_tr2, X_val2, X_ts2, y_tr2, y_val2, y_ts2 = \
        create_node_splits(GRAPHS_DIR, val_size, test_size, mode='by_img')
    assert X_tr.shape[0] == y_tr.shape[0] and X_val.shape[0] == y_val.shape[0] \
        and X_ts.shape[0] == y_ts.shape[0] and X_tr2.shape[0] == y_tr2.shape[0] \
        and X_val2.shape[0] == y_val2.shape[0] and X_ts2.shape[0] == y_ts2.shape[0] \
        and X_tr.shape[0] + X_val.shape[0] + X_ts.shape[0] == X_tr2.shape[0] + X_val2.shape[0] + X_ts2.shape[0] \
        and len(y_tr.shape) == 1 and len(y_val.shape) == 1 and len(y_ts.shape) == 1 \
        and len(y_tr2.shape) == 1 and len(y_val2.shape) == 1 and len(y_ts2.shape) == 1 \
        and y_tr.max() == 1 and y_val.max() == 1 and y_ts.max() == 1



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import shutil
import os
from argparse import Namespace
import pandas as pd

from tumourkit.utils.preprocessing import parse_path, create_dir
from tumourkit.postprocessing import join_graph_gt_main, join_hovprob_graph_main
from tumourkit import eval_class

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_DIR = parse_path(TEST_DIR)
JSON_DIR = TEST_DIR + 'json/'
GRAPHS_DIR = TEST_DIR + 'pipe_graphs/'
CENTROIDS_DIR = TEST_DIR + 'pipe_centroids/'
TMP_DIR = TEST_DIR + 'tmp/'


def test_hov_prob_pipe():
    create_dir(TMP_DIR)
    args = Namespace()
    args.graph_dir = GRAPHS_DIR
    args.centroids_dir = CENTROIDS_DIR
    args.output_dir = TMP_DIR
    join_graph_gt_main(args)
    args.node_dir = TMP_DIR
    args.save_file = TEST_DIR + 'tmp'
    args.by_img = False
    args.draw = False
    eval_class(args)
    args.json_dir = JSON_DIR
    args.graph_dir = TMP_DIR
    args.output_dir = TMP_DIR
    args.num_classes = 2
    args.enable_background = False
    join_hovprob_graph_main(args)
    args.node_dir = TMP_DIR
    args.save_file = TEST_DIR + 'tmp2'
    eval_class(args)
    res1 = pd.read_csv(TEST_DIR + 'tmp.csv')
    res2 = pd.read_csv(TEST_DIR + 'tmp2.csv')
    assert (abs(res1['Accuracy'] - res2['Accuracy']) < 0.01)[0]
    assert (abs(res1['F1-score'] - res2['F1-score']) < 0.01)[0]
    assert (abs(res1['Error percentage'] - res2['Error percentage']) < 0.01)[0]
    shutil.rmtree(TMP_DIR)
    os.remove(TEST_DIR + 'tmp.csv')
    os.remove(TEST_DIR + 'tmp2.csv')



import numpy as np
import torch.utils.data
from imgaug import augmenters as iaa
from ..misc.utils import cropping_center
from .augs import (
    add_to_brightness,
    add_to_contrast,
    add_to_hue,
    add_to_saturation,
    gaussian_blur,
    median_blur,
)


####
class FileLoader(torch.utils.data.Dataset):
    """Data Loader. Loads images from a file list and 
    performs augmentation with the albumentation library.
    After augmentation, horizontal and vertical maps are 
    generated.

    Args:
        file_list: list of filenames to load
        input_shape: shape of the input [h,w] - defined in config.py
        mask_shape: shape of the output [h,w] - defined in config.py
        mode: 'train' or 'valid'
        
    """

    # TODO: doc string

    def __init__(
        self,
        file_list,
        with_type=False,
        input_shape=None,
        mask_shape=None,
        mode="train",
        setup_augmentor=True,
        target_gen=None,
    ):
        assert input_shape is not None and mask_shape is not None
        self.mode = mode
        self.info_list = file_list
        self.with_type = with_type
        self.mask_shape = mask_shape
        self.input_shape = input_shape
        self.id = 0
        self.target_gen_func = target_gen[0]
        self.target_gen_kwargs = target_gen[1]
        if setup_augmentor:
            self.setup_augmentor(0, 0)
        return

    def setup_augmentor(self, worker_id, seed):
        self.augmentor = self.__get_augmentation(self.mode, seed)
        self.shape_augs = iaa.Sequential(self.augmentor[0])
        self.input_augs = iaa.Sequential(self.augmentor[1])
        self.id = self.id + worker_id
        return

    def __len__(self):
        return len(self.info_list)

    def __getitem__(self, idx):
        path = self.info_list[idx]
        data = np.load(path)

        # split stacked channel into image and label
        img = (data[..., :3]).astype("uint8")  # RGB images
        ann = (data[..., 3:]).astype("int32")  # instance ID map and type map

        if self.shape_augs is not None:
            shape_augs = self.shape_augs.to_deterministic()
            img = shape_augs.augment_image(img)
            ann = shape_augs.augment_image(ann)

        if self.input_augs is not None:
            input_augs = self.input_augs.to_deterministic()
            img = input_augs.augment_image(img)

        img = cropping_center(img, self.input_shape)
        feed_dict = {"img": img}

        inst_map = ann[..., 0]  # HW1 -> HW
        if self.with_type:
            type_map = (ann[..., 1]).copy()
            type_map = cropping_center(type_map, self.mask_shape)
            #type_map[type_map == 5] = 1  # merge neoplastic and non-neoplastic
            feed_dict["tp_map"] = type_map

        # TODO: document hard coded assumption about #input
        target_dict = self.target_gen_func(
            inst_map, self.mask_shape, **self.target_gen_kwargs
        )
        feed_dict.update(target_dict)

        return feed_dict

    def __get_augmentation(self, mode, rng):
        if mode == "train":
            shape_augs = [
                # * order = ``0`` -> ``cv2.INTER_NEAREST``
                # * order = ``1`` -> ``cv2.INTER_LINEAR``
                # * order = ``2`` -> ``cv2.INTER_CUBIC``
                # * order = ``3`` -> ``cv2.INTER_CUBIC``
                # * order = ``4`` -> ``cv2.INTER_CUBIC``
                # ! for pannuke v0, no rotation or translation, just flip to avoid mirror padding
                iaa.Affine(
                    # scale images to 80-120% of their size, individually per axis
                    scale={"x": (0.8, 1.2), "y": (0.8, 1.2)},
                    # translate by -A to +A percent (per axis)
                    translate_percent={"x": (-0.01, 0.01), "y": (-0.01, 0.01)},
                    shear=(-5, 5),  # shear by -5 to +5 degrees
                    rotate=(-179, 179),  # rotate by -179 to +179 degrees
                    order=0,  # use nearest neighbour
                    backend="cv2",  # opencv for fast processing
                    seed=rng,
                ),
                # set position to 'center' for center crop
                # else 'uniform' for random crop
                iaa.CropToFixedSize(
                    self.input_shape[0], self.input_shape[1], position="center"
                ),
                iaa.Fliplr(0.5, seed=rng),
                iaa.Flipud(0.5, seed=rng),
            ]

            input_augs = [
                iaa.OneOf(
                    [
                        iaa.Lambda(
                            seed=rng,
                            func_images=lambda *args: gaussian_blur(*args, max_ksize=3),
                        ),
                        iaa.Lambda(
                            seed=rng,
                            func_images=lambda *args: median_blur(*args, max_ksize=3),
                        ),
                        iaa.AdditiveGaussianNoise(
                            loc=0, scale=(0.0, 0.05 * 255), per_channel=0.5
                        ),
                    ]
                ),
                iaa.Sequential(
                    [
                        iaa.Lambda(
                            seed=rng,
                            func_images=lambda *args: add_to_hue(*args, range=(-8, 8)),
                        ),
                        iaa.Lambda(
                            seed=rng,
                            func_images=lambda *args: add_to_saturation(
                                *args, range=(-0.2, 0.2)
                            ),
                        ),
                        iaa.Lambda(
                            seed=rng,
                            func_images=lambda *args: add_to_brightness(
                                *args, range=(-26, 26)
                            ),
                        ),
                        iaa.Lambda(
                            seed=rng,
                            func_images=lambda *args: add_to_contrast(
                                *args, range=(0.75, 1.25)
                            ),
                        ),
                    ],
                    random_order=True,
                ),
            ]
        elif mode == "valid":
            shape_augs = [
                # set position to 'center' for center crop
                # else 'uniform' for random crop
                iaa.CropToFixedSize(
                    self.input_shape[0], self.input_shape[1], position="center"
                )
            ]
            input_augs = []

        return shape_augs, input_augs



import multiprocessing as mp
from concurrent.futures import FIRST_EXCEPTION, ProcessPoolExecutor, as_completed
from multiprocessing import Lock, Pool

mp.set_start_method("spawn", True)  # ! must be at top for VScode debugging
import glob
import logging
import os
import pathlib
import time
import cv2
import numpy as np
import torch.utils.data as data
import tqdm
from ..dataloader.infer_loader import SerializeArray
from ..misc.utils import (
    log_info,
    rm_n_mkdir,
)
from ..misc.wsi_handler import get_file_handler

from . import base

thread_lock = Lock()


####
def _init_worker_child(lock_):
    global lock
    lock = lock_


####
def _remove_inst(inst_map, remove_id_list):
    """Remove instances with id in remove_id_list.
    
    Args:
        inst_map: map of instances
        remove_id_list: list of ids to remove from inst_map
    """
    for inst_id in remove_id_list:
        inst_map[inst_map == inst_id] = 0
    return inst_map


####
def _get_patch_top_left_info(img_shape, input_size, output_size):
    """Get top left coordinate information of patches from original image.

    Args:
        img_shape: input image shape
        input_size: patch input shape
        output_size: patch output shape

    """
    in_out_diff = input_size - output_size
    nr_step = np.floor((img_shape - in_out_diff) / output_size) + 1
    last_output_coord = (in_out_diff // 2) + (nr_step) * output_size
    # generating subpatches index from orginal
    output_tl_y_list = np.arange(
        in_out_diff[0] // 2, last_output_coord[0], output_size[0], dtype=np.int32
    )
    output_tl_x_list = np.arange(
        in_out_diff[1] // 2, last_output_coord[1], output_size[1], dtype=np.int32
    )
    output_tl_y_list, output_tl_x_list = np.meshgrid(output_tl_y_list, output_tl_x_list)
    output_tl = np.stack(
        [output_tl_y_list.flatten(), output_tl_x_list.flatten()], axis=-1
    )
    input_tl = output_tl - in_out_diff // 2
    return input_tl, output_tl


#### all must be np.array
def _get_tile_info(img_shape, tile_shape, ambiguous_size=128):
    """Get information of tiles used for post processing.

    Args:
        img_shape: input image shape
        tile_shape: tile shape used for post processing
        ambiguous_size: used to define area at tile boundaries
    
    """
    # * get normal tiling set
    tile_grid_top_left, _ = _get_patch_top_left_info(img_shape, tile_shape, tile_shape)
    tile_grid_bot_right = []
    for idx in list(range(tile_grid_top_left.shape[0])):
        tile_tl = tile_grid_top_left[idx][:2]
        tile_br = tile_tl + tile_shape
        axis_sel = tile_br > img_shape
        tile_br[axis_sel] = img_shape[axis_sel]
        tile_grid_bot_right.append(tile_br)
    tile_grid_bot_right = np.array(tile_grid_bot_right)
    tile_grid = np.stack([tile_grid_top_left, tile_grid_bot_right], axis=1)
    tile_grid_x = np.unique(tile_grid_top_left[:, 1])
    tile_grid_y = np.unique(tile_grid_top_left[:, 0])
    # * get tiling set to fix vertical and horizontal boundary between tiles
    # for sanity, expand at boundary `ambiguous_size` to both side vertical and horizontal
    stack_coord = lambda x: np.stack([x[0].flatten(), x[1].flatten()], axis=-1)
    tile_boundary_x_top_left = np.meshgrid(
        tile_grid_y, tile_grid_x[1:] - ambiguous_size
    )
    tile_boundary_x_bot_right = np.meshgrid(
        tile_grid_y + tile_shape[0], tile_grid_x[1:] + ambiguous_size
    )
    tile_boundary_x_top_left = stack_coord(tile_boundary_x_top_left)
    tile_boundary_x_bot_right = stack_coord(tile_boundary_x_bot_right)
    tile_boundary_x = np.stack(
        [tile_boundary_x_top_left, tile_boundary_x_bot_right], axis=1
    )
    #
    tile_boundary_y_top_left = np.meshgrid(
        tile_grid_y[1:] - ambiguous_size, tile_grid_x
    )
    tile_boundary_y_bot_right = np.meshgrid(
        tile_grid_y[1:] + ambiguous_size, tile_grid_x + tile_shape[1]
    )
    tile_boundary_y_top_left = stack_coord(tile_boundary_y_top_left)
    tile_boundary_y_bot_right = stack_coord(tile_boundary_y_bot_right)
    tile_boundary_y = np.stack(
        [tile_boundary_y_top_left, tile_boundary_y_bot_right], axis=1
    )
    tile_boundary = np.concatenate([tile_boundary_x, tile_boundary_y], axis=0)
    # * get tiling set to fix the intersection of 4 tiles
    tile_cross_top_left = np.meshgrid(
        tile_grid_y[1:] - 2 * ambiguous_size, tile_grid_x[1:] - 2 * ambiguous_size
    )
    tile_cross_bot_right = np.meshgrid(
        tile_grid_y[1:] + 2 * ambiguous_size, tile_grid_x[1:] + 2 * ambiguous_size
    )
    tile_cross_top_left = stack_coord(tile_cross_top_left)
    tile_cross_bot_right = stack_coord(tile_cross_bot_right)
    tile_cross = np.stack([tile_cross_top_left, tile_cross_bot_right], axis=1)
    return tile_grid, tile_boundary, tile_cross


####
def _get_chunk_patch_info(
    img_shape, chunk_input_shape, patch_input_shape, patch_output_shape
):
    """Get chunk patch info. Here, chunk refers to tiles used during inference.

    Args:
        img_shape: input image shape
        chunk_input_shape: shape of tiles used for post processing
        patch_input_shape: input patch shape
        patch_output_shape: output patch shape

    """
    round_to_multiple = lambda x, y: np.floor(x / y) * y
    patch_diff_shape = patch_input_shape - patch_output_shape

    chunk_output_shape = chunk_input_shape - patch_diff_shape
    chunk_output_shape = round_to_multiple(
        chunk_output_shape, patch_output_shape
    ).astype(np.int64)
    chunk_input_shape = (chunk_output_shape + patch_diff_shape).astype(np.int64)

    patch_input_tl_list, _ = _get_patch_top_left_info(
        img_shape, patch_input_shape, patch_output_shape
    )
    patch_input_br_list = patch_input_tl_list + patch_input_shape
    patch_output_tl_list = patch_input_tl_list + patch_diff_shape
    patch_output_br_list = patch_output_tl_list + patch_output_shape
    patch_info_list = np.stack(
        [
            np.stack([patch_input_tl_list, patch_input_br_list], axis=1),
            np.stack([patch_output_tl_list, patch_output_br_list], axis=1),
        ],
        axis=1,
    )

    chunk_input_tl_list, _ = _get_patch_top_left_info(
        img_shape, chunk_input_shape, chunk_output_shape
    )
    chunk_input_br_list = chunk_input_tl_list + chunk_input_shape
    # * correct the coord so it stay within source image
    y_sel = np.nonzero(chunk_input_br_list[:, 0] > img_shape[0])[0]
    x_sel = np.nonzero(chunk_input_br_list[:, 1] > img_shape[1])[0]
    chunk_input_br_list[y_sel, 0] = (
        img_shape[0] - patch_diff_shape[0]
    ) - chunk_input_tl_list[y_sel, 0]
    chunk_input_br_list[x_sel, 1] = (
        img_shape[1] - patch_diff_shape[1]
    ) - chunk_input_tl_list[x_sel, 1]
    chunk_input_br_list[y_sel, 0] = round_to_multiple(
        chunk_input_br_list[y_sel, 0], patch_output_shape[0]
    )
    chunk_input_br_list[x_sel, 1] = round_to_multiple(
        chunk_input_br_list[x_sel, 1], patch_output_shape[1]
    )
    chunk_input_br_list[y_sel, 0] += chunk_input_tl_list[y_sel, 0] + patch_diff_shape[0]
    chunk_input_br_list[x_sel, 1] += chunk_input_tl_list[x_sel, 1] + patch_diff_shape[1]
    chunk_output_tl_list = chunk_input_tl_list + patch_diff_shape // 2
    chunk_output_br_list = chunk_input_br_list - patch_diff_shape // 2  # may off pixels
    chunk_info_list = np.stack(
        [
            np.stack([chunk_input_tl_list, chunk_input_br_list], axis=1),
            np.stack([chunk_output_tl_list, chunk_output_br_list], axis=1),
        ],
        axis=1,
    )

    return chunk_info_list, patch_info_list


####
def _post_proc_para_wrapper(pred_map_mmap_path, tile_info, func, func_kwargs):
    """Wrapper for parallel post processing."""
    idx, tile_tl, tile_br = tile_info
    wsi_pred_map_ptr = np.load(pred_map_mmap_path, mmap_mode="r")
    tile_pred_map = wsi_pred_map_ptr[tile_tl[0] : tile_br[0], tile_tl[1] : tile_br[1]]
    tile_pred_map = np.array(tile_pred_map)  # from mmap to ram
    return func(tile_pred_map, **func_kwargs), tile_info


####
def _assemble_and_flush(wsi_pred_map_mmap_path, chunk_info, patch_output_list):
    """Assemble the results. Write to newly created holder for this wsi"""
    wsi_pred_map_ptr = np.load(wsi_pred_map_mmap_path, mmap_mode="r+")
    chunk_pred_map = wsi_pred_map_ptr[
        chunk_info[1][0][0] : chunk_info[1][1][0],
        chunk_info[1][0][1] : chunk_info[1][1][1],
    ]
    if patch_output_list is None:
        # chunk_pred_map[:] = 0 # zero flush when there is no-results
        # print(chunk_info.flatten(), 'flush 0')
        return

    for pinfo in patch_output_list:
        pcoord, pdata = pinfo
        pdata = np.squeeze(pdata)
        pcoord = np.squeeze(pcoord)[:2]
        chunk_pred_map[
            pcoord[0] : pcoord[0] + pdata.shape[0],
            pcoord[1] : pcoord[1] + pdata.shape[1],
        ] = pdata
    # print(chunk_info.flatten(), 'pass')
    return


####
class InferManager(base.InferManager):
    def __run_model(self, patch_top_left_list, pbar_desc):
        # TODO: the cost of creating dataloader may not be cheap ?
        dataset = SerializeArray(
            "%s/cache_chunk.npy" % self.cache_path,
            patch_top_left_list,
            self.patch_input_shape,
        )

        dataloader = data.DataLoader(
            dataset,
            num_workers=self.nr_inference_workers,
            batch_size=self.batch_size,
            drop_last=False,
        )

        pbar = tqdm.tqdm(
            desc=pbar_desc,
            leave=True,
            total=int(len(dataloader)),
            ncols=80,
            ascii=True,
            position=0,
        )

        # run inference on input patches
        accumulated_patch_output = []
        for batch_idx, batch_data in enumerate(dataloader):
            sample_data_list, sample_info_list = batch_data
            sample_output_list = self.run_step(sample_data_list)
            sample_info_list = sample_info_list.numpy()
            curr_batch_size = sample_output_list.shape[0]
            sample_output_list = np.split(sample_output_list, curr_batch_size, axis=0)
            sample_info_list = np.split(sample_info_list, curr_batch_size, axis=0)
            sample_output_list = list(zip(sample_info_list, sample_output_list))
            accumulated_patch_output.extend(sample_output_list)
            pbar.update()
        pbar.close()
        return accumulated_patch_output

    def __select_valid_patches(self, patch_info_list, has_output_info=True):
        """Select valid patches from the list of input patch information.

        Args:
            patch_info_list: patch input coordinate information
            has_output_info: whether output information is given
        
        """
        down_sample_ratio = self.wsi_mask.shape[0] / self.wsi_proc_shape[0]
        selected_indices = []
        for idx in range(patch_info_list.shape[0]):
            patch_info = patch_info_list[idx]
            patch_info = np.squeeze(patch_info)
            # get the box at corresponding mag of the mask
            if has_output_info:
                output_bbox = patch_info[1] * down_sample_ratio
            else:
                output_bbox = patch_info * down_sample_ratio
            output_bbox = np.rint(output_bbox).astype(np.int64)
            # coord of the output of the patch (i.e center regions)
            output_roi = self.wsi_mask[
                output_bbox[0][0] : output_bbox[1][0],
                output_bbox[0][1] : output_bbox[1][1],
            ]
            if np.sum(output_roi) > 0:
                selected_indices.append(idx)
        sub_patch_info_list = patch_info_list[selected_indices]
        return sub_patch_info_list

    def __get_raw_prediction(self, chunk_info_list, patch_info_list):
        """Process input tiles (called chunks for inference) with HoVer-Net.

        Args:
            chunk_info_list: list of inference tile coordinate information
            patch_info_list: list of patch coordinate information
        
        """
        # 1 dedicated thread just to write results back to disk
        proc_pool = Pool(processes=1)
        wsi_pred_map_mmap_path = "%s/pred_map.npy" % self.cache_path

        masking = lambda x, a, b: (a <= x) & (x <= b)
        for idx in range(0, chunk_info_list.shape[0]):
            chunk_info = chunk_info_list[idx]
            # select patch basing on top left coordinate of input
            start_coord = chunk_info[0, 0]
            end_coord = chunk_info[0, 1] - self.patch_input_shape
            selection = masking(
                patch_info_list[:, 0, 0, 0], start_coord[0], end_coord[0]
            ) & masking(patch_info_list[:, 0, 0, 1], start_coord[1], end_coord[1])
            chunk_patch_info_list = np.array(
                patch_info_list[selection]
            )  # * do we need copy ?

            # further select only the patches within the provided mask
            chunk_patch_info_list = self.__select_valid_patches(chunk_patch_info_list)

            # there no valid patches, so flush 0 and skip
            if chunk_patch_info_list.shape[0] == 0:
                proc_pool.apply_async(
                    _assemble_and_flush, args=(wsi_pred_map_mmap_path, chunk_info, None)
                )
                continue

            # shift the coordinare from wrt slide to wrt chunk
            chunk_patch_info_list -= chunk_info[:, 0]
            chunk_data = self.wsi_handler.read_region(
                chunk_info[0][0][::-1], (chunk_info[0][1] - chunk_info[0][0])[::-1]
            )
            chunk_data = np.array(chunk_data)[..., :3]
            np.save("%s/cache_chunk.npy" % self.cache_path, chunk_data)

            pbar_desc = "Process Chunk %d/%d" % (idx, chunk_info_list.shape[0])
            patch_output_list = self.__run_model(
                chunk_patch_info_list[:, 0, 0], pbar_desc
            )

            proc_pool.apply_async(
                _assemble_and_flush,
                args=(wsi_pred_map_mmap_path, chunk_info, patch_output_list),
            )
        proc_pool.close()
        proc_pool.join()
        return

    def __dispatch_post_processing(self, tile_info_list, callback):
        """Post processing initialisation."""
        proc_pool = None
        if self.nr_post_proc_workers > 0:
            proc_pool = ProcessPoolExecutor(self.nr_post_proc_workers)

        future_list = []
        wsi_pred_map_mmap_path = "%s/pred_map.npy" % self.cache_path
        for idx in list(range(tile_info_list.shape[0])):
            tile_tl = tile_info_list[idx][0]
            tile_br = tile_info_list[idx][1]

            tile_info = (idx, tile_tl, tile_br)
            func_kwargs = {
                "nr_types": self.method["model_args"]["nr_types"],
                "return_centroids": True,
            }

            # TODO: standarize protocol
            if proc_pool is not None:
                proc_future = proc_pool.submit(
                    _post_proc_para_wrapper,
                    wsi_pred_map_mmap_path,
                    tile_info,
                    self.post_proc_func,
                    func_kwargs,
                )
                # ! manually poll future and call callback later as there is no guarantee
                # ! that the callback is called from main thread
                future_list.append(proc_future)
            else:
                results = _post_proc_para_wrapper(
                    wsi_pred_map_mmap_path, tile_info, self.post_proc_func, func_kwargs
                )
                callback(results)
        if proc_pool is not None:
            silent_crash = False
            # loop over all to check state a.k.a polling
            for future in as_completed(future_list):
                # ! silent crash, cancel all and raise error
                if future.exception() is not None:
                    silent_crash = True
                    # ! cancel somehow leads to cascade error later
                    # ! so just poll it then crash once all future
                    # ! acquired for now
                    # for future in future_list:
                    #     future.cancel()
                    # break
                else:
                    callback(future.result())
            assert not silent_crash
        return

    def _parse_args(self, run_args):
        """Parse command line arguments and set as instance variables."""
        for variable, value in run_args.items():
            self.__setattr__(variable, value)
        # to tuple
        self.chunk_shape = [self.chunk_shape, self.chunk_shape]
        self.tile_shape = [self.tile_shape, self.tile_shape]
        self.patch_input_shape = [self.patch_input_shape, self.patch_input_shape]
        self.patch_output_shape = [self.patch_output_shape, self.patch_output_shape]
        return

    def process_single_file(self, wsi_path, msk_path, output_dir):
        """Process a single whole-slide image and save the results.

        Args:
            wsi_path: path to input whole-slide image
            msk_path: path to input mask. If not supplied, mask will be automatically generated.
            output_dir: path where output will be saved

        """
        # TODO: customize universal file handler to sync the protocol
        ambiguous_size = self.ambiguous_size
        tile_shape = (np.array(self.tile_shape)).astype(np.int64)
        chunk_input_shape = np.array(self.chunk_shape)
        patch_input_shape = np.array(self.patch_input_shape)
        patch_output_shape = np.array(self.patch_output_shape)

        path_obj = pathlib.Path(wsi_path)
        wsi_ext = path_obj.suffix
        wsi_name = path_obj.stem

        start = time.perf_counter()
        self.wsi_handler = get_file_handler(wsi_path, backend=wsi_ext)
        self.wsi_proc_shape = self.wsi_handler.get_dimensions(self.proc_mag)
        self.wsi_handler.prepare_reading(
            read_mag=self.proc_mag, cache_path="%s/src_wsi.npy" % self.cache_path
        )
        self.wsi_proc_shape = np.array(self.wsi_proc_shape[::-1])  # to Y, X

        if msk_path is not None and os.path.isfile(msk_path):
            self.wsi_mask = cv2.imread(msk_path)
            self.wsi_mask = cv2.cvtColor(self.wsi_mask, cv2.COLOR_BGR2GRAY)
            self.wsi_mask[self.wsi_mask > 0] = 1
        else:
            log_info(
                "WARNING: No mask found, generating mask via thresholding at 1.25x!"
            )

            from skimage import morphology

            # simple method to extract tissue regions using intensity thresholding and morphological operations
            def simple_get_mask():
                scaled_wsi_mag = 1.25  # ! hard coded
                wsi_thumb_rgb = self.wsi_handler.get_full_img(read_mag=scaled_wsi_mag)
                gray = cv2.cvtColor(wsi_thumb_rgb, cv2.COLOR_RGB2GRAY)
                _, mask = cv2.threshold(gray, 0, 255, cv2.THRESH_OTSU)
                mask = morphology.remove_small_objects(
                    mask == 0, min_size=16 * 16, connectivity=2
                )
                mask = morphology.remove_small_holes(mask, area_threshold=128 * 128)
                mask = morphology.binary_dilation(mask, morphology.disk(16))
                return mask

            self.wsi_mask = np.array(simple_get_mask() > 0, dtype=np.uint8)
        if np.sum(self.wsi_mask) == 0:
            log_info("Skip due to empty mask!")
            return
        if self.save_mask:
            cv2.imwrite("%s/mask/%s.png" % (output_dir, wsi_name), self.wsi_mask * 255)
        if self.save_thumb:
            wsi_thumb_rgb = self.wsi_handler.get_full_img(read_mag=1.25)
            cv2.imwrite(
                "%s/thumb/%s.png" % (output_dir, wsi_name),
                cv2.cvtColor(wsi_thumb_rgb, cv2.COLOR_RGB2BGR),
            )

        # * declare holder for output
        # create a memory-mapped .npy file with the predefined dimensions and dtype
        # TODO: dynamicalize this, retrieve from model?
        out_ch = 3 if self.method["model_args"]["nr_types"] is None else 4
        self.wsi_inst_info = {}
        # TODO: option to use entire RAM if users have too much available, would be faster than mmap
        self.wsi_inst_map = np.lib.format.open_memmap(
            "%s/pred_inst.npy" % self.cache_path,
            mode="w+",
            shape=tuple(self.wsi_proc_shape),
            dtype=np.int32,
        )
        # self.wsi_inst_map[:] = 0 # flush fill

        # warning, the value within this is uninitialized
        self.wsi_pred_map = np.lib.format.open_memmap(
            "%s/pred_map.npy" % self.cache_path,
            mode="w+",
            shape=tuple(self.wsi_proc_shape) + (out_ch,),
            dtype=np.float32,
        )
        # ! for debug
        # self.wsi_pred_map = np.load('%s/pred_map.npy' % self.cache_path, mmap_mode='r')
        end = time.perf_counter()
        log_info("Preparing Input Output Placement: {0}".format(end - start))

        # * raw prediction
        start = time.perf_counter()
        chunk_info_list, patch_info_list = _get_chunk_patch_info(
            self.wsi_proc_shape,
            chunk_input_shape,
            patch_input_shape,
            patch_output_shape,
        )

        # get the raw prediction of HoVer-Net, given info of inference tiles and patches
        self.__get_raw_prediction(chunk_info_list, patch_info_list)
        end = time.perf_counter()
        log_info("Inference Time: {0}".format(end - start))

        # TODO: deal with error banding
        ##### * post processing
        ##### * done in 3 stages to ensure that nuclei at the boundaries are dealt with accordingly
        start = time.perf_counter()
        tile_coord_set = _get_tile_info(self.wsi_proc_shape, tile_shape, ambiguous_size)
        # 3 sets of patches are extracted and are dealt with differently
        # tile_grid_info: central region of post processing tiles
        # tile_boundary_info: boundary region of post processing tiles
        # tile_cross_info: region at corners of post processing tiles
        tile_grid_info, tile_boundary_info, tile_cross_info = tile_coord_set
        tile_grid_info = self.__select_valid_patches(tile_grid_info, False)
        tile_boundary_info = self.__select_valid_patches(tile_boundary_info, False)
        tile_cross_info = self.__select_valid_patches(tile_cross_info, False)

        ####################### * Callback can only receive 1 arg
        def post_proc_normal_tile_callback(args):
            results, pos_args = args
            run_idx, tile_tl, tile_br = pos_args
            pred_inst, inst_info_dict = results

            if len(inst_info_dict) == 0:
                pbar.update()  # external
                return  # when there is nothing to do

            top_left = pos_args[1][::-1]

            # ! WARNING:
            # ! inst ID may not be contiguous,
            # ! hence must use max as safeguard

            wsi_max_id = 0
            if len(self.wsi_inst_info) > 0:
                wsi_max_id = max(self.wsi_inst_info.keys())
            for inst_id, inst_info in inst_info_dict.items():
                # now correct the coordinate wrt to wsi
                inst_info["bbox"] += top_left
                inst_info["contour"] += top_left
                inst_info["centroid"] += top_left
                self.wsi_inst_info[inst_id + wsi_max_id] = inst_info
            pred_inst[pred_inst > 0] += wsi_max_id
            self.wsi_inst_map[
                tile_tl[0] : tile_br[0], tile_tl[1] : tile_br[1]
            ] = pred_inst

            pbar.update()  # external
            return

        ####################### * Callback can only receive 1 arg
        def post_proc_fixing_tile_callback(args):
            results, pos_args = args
            run_idx, tile_tl, tile_br = pos_args
            pred_inst, inst_info_dict = results

            if len(inst_info_dict) == 0:
                pbar.update()  # external
                return  # when there is nothing to do

            top_left = pos_args[1][::-1]

            # for fixing the boundary, keep all nuclei split at boundary (i.e within unambigous region)
            # of the existing prediction map, and replace all nuclei within the region with newly predicted

            # ! WARNING:
            # ! inst ID may not be contiguous,
            # ! hence must use max as safeguard

            # ! must get before the removal happened
            wsi_max_id = 0
            if len(self.wsi_inst_info) > 0:
                wsi_max_id = max(self.wsi_inst_info.keys())

            # * exclude ambiguous out from old prediction map
            # check 1 pix of 4 edges to find nuclei split at boundary
            roi_inst = self.wsi_inst_map[
                tile_tl[0] : tile_br[0], tile_tl[1] : tile_br[1]
            ]
            roi_inst = np.copy(roi_inst)
            roi_edge = np.concatenate(
                [roi_inst[[0, -1], :].flatten(), roi_inst[:, [0, -1]].flatten()]
            )
            roi_boundary_inst_list = np.unique(roi_edge)[1:]  # exclude background
            roi_inner_inst_list = np.unique(roi_inst)[1:]
            roi_inner_inst_list = np.setdiff1d(
                roi_inner_inst_list, roi_boundary_inst_list, assume_unique=True
            )
            roi_inst = _remove_inst(roi_inst, roi_inner_inst_list)
            self.wsi_inst_map[
                tile_tl[0] : tile_br[0], tile_tl[1] : tile_br[1]
            ] = roi_inst
            for inst_id in roi_inner_inst_list:
                self.wsi_inst_info.pop(inst_id, None)

            # * exclude unambiguous out from new prediction map
            # check 1 pix of 4 edges to find nuclei split at boundary
            roi_edge = pred_inst[roi_inst > 0]  # remove all overlap
            boundary_inst_list = np.unique(roi_edge)  # no background to exclude
            inner_inst_list = np.unique(pred_inst)[1:]
            inner_inst_list = np.setdiff1d(
                inner_inst_list, boundary_inst_list, assume_unique=True
            )
            pred_inst = _remove_inst(pred_inst, boundary_inst_list)

            # * proceed to overwrite
            for inst_id in inner_inst_list:
                # ! happen because we alrd skip thoses with wrong
                # ! contour (<3 points) within the postproc, so
                # ! sanity gate here
                if inst_id not in inst_info_dict:
                    log_info("Nuclei id=%d not in saved dict WRN1." % inst_id)
                    continue
                inst_info = inst_info_dict[inst_id]
                # now correct the coordinate wrt to wsi
                inst_info["bbox"] += top_left
                inst_info["contour"] += top_left
                inst_info["centroid"] += top_left
                self.wsi_inst_info[inst_id + wsi_max_id] = inst_info
            pred_inst[pred_inst > 0] += wsi_max_id
            pred_inst = roi_inst + pred_inst
            self.wsi_inst_map[
                tile_tl[0] : tile_br[0], tile_tl[1] : tile_br[1]
            ] = pred_inst

            pbar.update()  # external
            return

        #######################
        pbar_creator = lambda x, y: tqdm.tqdm(
            desc=y, leave=True, total=int(len(x)), ncols=80, ascii=True, position=0
        )
        pbar = pbar_creator(tile_grid_info, "Post Proc Phase 1")
        # * must be in sequential ordering
        self.__dispatch_post_processing(tile_grid_info, post_proc_normal_tile_callback)
        pbar.close()

        pbar = pbar_creator(tile_boundary_info, "Post Proc Phase 2")
        self.__dispatch_post_processing(
            tile_boundary_info, post_proc_fixing_tile_callback
        )
        pbar.close()

        pbar = pbar_creator(tile_cross_info, "Post Proc Phase 3")
        self.__dispatch_post_processing(tile_cross_info, post_proc_fixing_tile_callback)
        pbar.close()

        end = time.perf_counter()
        log_info("Total Post Proc Time: {0}".format(end - start))

        # ! cant possibly save the inst map at high res, too large
        start = time.perf_counter()
        if self.save_mask or self.save_thumb:
            json_path = "%s/json/%s.json" % (output_dir, wsi_name)
        else:
            json_path = "%s/%s.json" % (output_dir, wsi_name)
        self.__save_json(json_path, self.wsi_inst_info, mag=self.proc_mag)
        end = time.perf_counter()
        log_info("Save Time: {0}".format(end - start))

    def process_wsi_list(self, run_args):
        """Process a list of whole-slide images.

        Args:
            run_args: arguments as defined in run_infer.py
        
        """
        self._parse_args(run_args)

        if not os.path.exists(self.cache_path):
            rm_n_mkdir(self.cache_path)

        if not os.path.exists(self.output_dir + "/json/"):
            rm_n_mkdir(self.output_dir + "/json/")
        if self.save_thumb:
            if not os.path.exists(self.output_dir + "/thumb/"):
                rm_n_mkdir(self.output_dir + "/thumb/")
        if self.save_mask:
            if not os.path.exists(self.output_dir + "/mask/"):
                rm_n_mkdir(self.output_dir + "/mask/")

        wsi_path_list = glob.glob(self.input_dir + "/*")
        wsi_path_list.sort()  # ensure ordering
        for wsi_path in wsi_path_list[:]:
            wsi_base_name = pathlib.Path(wsi_path).stem
            msk_path = "%s/%s.png" % (self.input_mask_dir, wsi_base_name)
            if self.save_thumb or self.save_mask:
                output_file = "%s/json/%s.json" % (self.output_dir, wsi_base_name)
            else:
                output_file = "%s/%s.json" % (self.output_dir, wsi_base_name)
            if os.path.exists(output_file):
                log_info("Skip: %s" % wsi_base_name)
                continue
            try:
                log_info("Process: %s" % wsi_base_name)
                self.process_single_file(wsi_path, msk_path, self.output_dir)
                log_info("Finish")
            except:
                logging.exception("Crash")
        rm_n_mkdir(self.cache_path)  # clean up all cache
        return



import importlib
from typing import Optional

class Config(object):
    """Configuration file."""

    def __init__(self, shape: str, log_dir: str, train_dir: str, valid_dir: str, pretrained_path: str, use_cpu: bool, num_classes: Optional[int] = 2):
        self.seed = 10
        self.logging = True
        self.debug = False

        model_name = "hovernet"
        model_mode = "original" # choose either `original` or `fast`

        if model_mode not in ["original", "fast"]:
            raise Exception("Must use either `original` or `fast` as model mode")

        nr_type = num_classes + 1 # number of nuclear types (including background)

        # whether to predict the nuclear type, availability depending on dataset!
        self.type_classification = True

        if shape == '270':
            aug_shape = [540, 540] # patch shape used during augmentation (larger patch may have less border artefacts)
            act_shape = [270, 270] # patch shape used as input to network - central crop performed after augmentation
            out_shape = [80, 80]   # patch shape at output of network
        elif shape == '518':
            aug_shape = [540, 540] # patch shape used during augmentation (larger patch may have less border artefacts)
            act_shape = [518, 518] # patch shape used as input to network - central crop performed after augmentation
            out_shape = [328, 328] # patch shape at output of network
        else:
            assert False, 'Wrong shape indicated in config file.'

        self.log_dir = log_dir # where checkpoints will be saved

        # paths to training and validation patches
        self.train_dir_list = [train_dir]
        self.valid_dir_list = [valid_dir]

        self.shape_info = {
            "train": {"input_shape": act_shape, "mask_shape": out_shape,},
            "valid": {"input_shape": act_shape, "mask_shape": out_shape,},
        }

        module = importlib.import_module(
            ".segmentation.hovernet.models.%s.opt" % model_name, package="tumourkit"
        )
        self.model_config = module.get_config(nr_type, model_mode)
        self.model_config["phase_list"][0]["run_info"]["net"]["pretrained"] = pretrained_path
        self.model_config["phase_list"][0]["run_info"]["net"]["use_cpu"] = use_cpu
        self.model_config["phase_list"][1]["run_info"]["net"]["use_cpu"] = use_cpu



"""
Script to visualize the result from extract_tensorboard.
It provides several plots about how the hyperparameter affects the metrics.


Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
import pandas as pd
import os
import altair as alt


def plot_vis(values: pd.DataFrame, path: str) -> None:
    """
    Creates interesting altair plots.

    :param values: Dataframe with hyperparams in index and metrics in columns.
    :type values: pd.DataFrame
    :param path: Folder where to save plots, must exists already.
    :type path: str
    """
    metrics = values['metric'].unique()
    archs = values['arch'].unique()
    for arch in archs:
        os.makedirs(os.path.join(path, 'batch-norm', arch), exist_ok=True)
        os.makedirs(os.path.join(path, 'dropout', arch), exist_ok=True)
        for metric in metrics:
            # Is batch normalization useful?
            x_var = 'layers'
            y_var = 'max(value)'
            box_var = 'value'
            color_var = 'bn'
            data = values[(values.split == 'validation') & (values.metric == metric) & (values.arch == arch)]
            lines = alt.Chart(data).mark_line().encode(
                x=alt.X(x_var, title='Number of layers'),
                y=alt.Y(y_var, title=metric),
                color=alt.Color(color_var, legend=alt.Legend(title="BatchNorm")),
                tooltip=[x_var, y_var, color_var]
            )
            lines.save(os.path.join(path, 'batch-norm', arch, metric + '-' + arch + '-bn-line.svg'))
            lines.save(os.path.join(path, 'batch-norm', arch, metric + '-' + arch + '-bn-line.png'), scale_factor=2.0)
            # How does dropout affect the result?
            x_var = 'dropout'
            boxplot = alt.Chart(data).mark_boxplot().encode(
                x=alt.X(x_var, title='Percentage of dropout'),
                y=alt.Y(box_var, title=metric),
            )
            boxplot.save(os.path.join(path, 'dropout', arch, metric + '-' + arch + '-drop-box.svg'))
            boxplot.save(os.path.join(path, 'dropout', arch, metric + '-' + arch + '-drop-box.png'), scale_factor=2.0)

    return


def _create_parser():
    parser = argparse.ArgumentParser('Plot tensorboard logs in a better way.')
    parser.add_argument('--input-path', type=str, required=True, help='Path to the .csv from extract_tensorboard.')
    parser.add_argument('--output-dir', type=str, required=True, help='Path to folder where to save plots.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    os.makedirs(args.output_dir, exist_ok=True)
    values = pd.read_csv(args.input_path)
    plot_vis(values, args.output_dir)


import multiprocessing
from multiprocessing import Lock, Pool
multiprocessing.set_start_method("spawn", True)  # ! must be at top for VScode debugging
import glob
import math
import pathlib
import re
import sys
from concurrent.futures import FIRST_EXCEPTION, ProcessPoolExecutor, as_completed
import cv2
import numpy as np
import psutil
import scipy.io as sio
import torch.utils.data as data
import tqdm
from ..dataloader.infer_loader import SerializeFileList
from ..misc.utils import (
    log_info,
    rm_n_mkdir,
)
from ..misc.viz_utils import visualize_instances_dict
from . import base


####
def _prepare_patching(img, window_size, mask_size, return_src_top_corner=False):
    """Prepare patch information for tile processing.
    
    Args:
        img: original input image
        window_size: input patch size
        mask_size: output patch size
        return_src_top_corner: whether to return coordiante information for top left corner of img
        
    """

    win_size = window_size
    msk_size = step_size = mask_size

    def get_last_steps(length, msk_size, step_size):
        nr_step = math.ceil((length - msk_size) / step_size)
        last_step = (nr_step + 1) * step_size
        return int(last_step), int(nr_step + 1)

    im_h = img.shape[0]
    im_w = img.shape[1]

    last_h, _ = get_last_steps(im_h, msk_size, step_size)
    last_w, _ = get_last_steps(im_w, msk_size, step_size)

    diff = win_size - step_size
    padt = padl = diff // 2
    padb = last_h + win_size - im_h
    padr = last_w + win_size - im_w

    img = np.lib.pad(img, ((padt, padb), (padl, padr), (0, 0)), "reflect")

    # generating subpatches index from orginal
    coord_y = np.arange(0, last_h, step_size, dtype=np.int32)
    coord_x = np.arange(0, last_w, step_size, dtype=np.int32)
    row_idx = np.arange(0, coord_y.shape[0], dtype=np.int32)
    col_idx = np.arange(0, coord_x.shape[0], dtype=np.int32)
    coord_y, coord_x = np.meshgrid(coord_y, coord_x)
    row_idx, col_idx = np.meshgrid(row_idx, col_idx)
    coord_y = coord_y.flatten()
    coord_x = coord_x.flatten()
    row_idx = row_idx.flatten()
    col_idx = col_idx.flatten()
    #
    patch_info = np.stack([coord_y, coord_x, row_idx, col_idx], axis=-1)
    if not return_src_top_corner:
        return img, patch_info
    else:
        return img, patch_info, [padt, padl]


####
def _post_process_patches(
    post_proc_func, post_proc_kwargs, patch_info, image_info, overlay_kwargs,
):
    """Apply post processing to patches.
    
    Args:
        post_proc_func: post processing function to use
        post_proc_kwargs: keyword arguments used in post processing function
        patch_info: patch data and associated information
        image_info: input image data and associated information
        overlay_kwargs: overlay keyword arguments

    """
    # re-assemble the prediction, sort according to the patch location within the original image
    patch_info = sorted(patch_info, key=lambda x: [x[0][0], x[0][1]])
    patch_info, patch_data = zip(*patch_info)

    src_shape = image_info["src_shape"]
    src_image = image_info["src_image"]

    patch_shape = np.squeeze(patch_data[0]).shape
    ch = 1 if len(patch_shape) == 2 else patch_shape[-1]
    axes = [0, 2, 1, 3, 4] if ch != 1 else [0, 2, 1, 3]

    nr_row = max([x[2] for x in patch_info]) + 1
    nr_col = max([x[3] for x in patch_info]) + 1
    pred_map = np.concatenate(patch_data, axis=0)
    pred_map = np.reshape(pred_map, (nr_row, nr_col) + patch_shape)
    pred_map = np.transpose(pred_map, axes)
    pred_map = np.reshape(
        pred_map, (patch_shape[0] * nr_row, patch_shape[1] * nr_col, ch)
    )
    # crop back to original shape
    pred_map = np.squeeze(pred_map[: src_shape[0], : src_shape[1]])

    # * Implicit protocol
    # * a prediction map with instance of ID 1-N
    # * and a dict contain the instance info, access via its ID
    # * each instance may have type
    pred_inst, inst_info_dict = post_proc_func(pred_map, **post_proc_kwargs)

    overlaid_img = visualize_instances_dict(
        src_image.copy(), inst_info_dict, **overlay_kwargs
    )

    return image_info["name"], pred_map, pred_inst, inst_info_dict, overlaid_img


class InferManager(base.InferManager):
    """Run inference on tiles."""

    ####
    def process_file_list(self, run_args):
        """
        Process a single image tile < 5000x5000 in size.
        """
        for variable, value in run_args.items():
            self.__setattr__(variable, value)
        assert self.mem_usage < 1.0 and self.mem_usage > 0.0

        # * depend on the number of samples and their size, this may be less efficient
        patterning = lambda x: re.sub("([\[\]])", "[\\1]", x)
        file_path_list = glob.glob(patterning("%s/*" % self.input_dir))
        file_path_list.sort()  # ensure same order
        assert len(file_path_list) > 0, 'Not Detected Any Files From Path'
        
        rm_n_mkdir(self.output_dir + '/json/')
        rm_n_mkdir(self.output_dir + '/mat/')
        rm_n_mkdir(self.output_dir + '/overlay/')
        if self.save_qupath:
            rm_n_mkdir(self.output_dir + "/qupath/")

        def proc_callback(results):
            """Post processing callback.
            
            Output format is implicit assumption, taken from `_post_process_patches`

            """
            img_name, pred_map, pred_inst, inst_info_dict, overlaid_img = results

            nuc_val_list = list(inst_info_dict.values())
            # need singleton to make matlab happy
            nuc_uid_list = np.array(list(inst_info_dict.keys()))[:,None]
            nuc_type_list = np.array([v["type"] for v in nuc_val_list])[:,None]
            nuc_coms_list = np.array([v["centroid"] for v in nuc_val_list])

            mat_dict = {
                "inst_map" : pred_inst,
                "inst_uid" : nuc_uid_list,
                "inst_type": nuc_type_list,
                "inst_centroid": nuc_coms_list
            }
            if self.nr_types is None: # matlab does not have None type array
                mat_dict.pop("inst_type", None) 

            if self.save_raw_map:
                mat_dict["raw_map"] = pred_map
            save_path = "%s/mat/%s.mat" % (self.output_dir, img_name)
            sio.savemat(save_path, mat_dict)

            save_path = "%s/overlay/%s.png" % (self.output_dir, img_name)
            cv2.imwrite(save_path, cv2.cvtColor(overlaid_img, cv2.COLOR_RGB2BGR))

            save_path = "%s/json/%s.json" % (self.output_dir, img_name)
            self.__save_json(save_path, inst_info_dict, None)
            return img_name

        def detach_items_of_uid(items_list, uid, nr_expected_items):
            item_counter = 0
            detached_items_list = []
            remained_items_list = []
            while True:
                pinfo, pdata = items_list.pop(0)
                pinfo = np.squeeze(pinfo)
                if pinfo[-1] == uid:
                    detached_items_list.append([pinfo, pdata])
                    item_counter += 1
                else:
                    remained_items_list.append([pinfo, pdata])
                if item_counter == nr_expected_items:
                    break
            # do this to ensure the ordering
            remained_items_list = remained_items_list + items_list
            return detached_items_list, remained_items_list

        proc_pool = None
        if self.nr_post_proc_workers > 0:
            proc_pool = ProcessPoolExecutor(self.nr_post_proc_workers)

        while len(file_path_list) > 0:

            hardware_stats = psutil.virtual_memory()
            available_ram = getattr(hardware_stats, "available")
            available_ram = int(available_ram * self.mem_usage)
            # available_ram >> 20 for MB, >> 30 for GB

            # TODO: this portion looks clunky but seems hard to detach into separate func

            # * caching N-files into memory such that their expected (total) memory usage
            # * does not exceed the designated percentage of currently available memory
            # * the expected memory is a factor w.r.t original input file size and
            # * must be manually provided
            file_idx = 0
            use_path_list = []
            cache_image_list = []
            cache_patch_info_list = []
            cache_image_info_list = []
            while len(file_path_list) > 0:
                file_path = file_path_list.pop(0)

                img = cv2.imread(file_path)
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                src_shape = img.shape

                img, patch_info, top_corner = _prepare_patching(
                    img, self.patch_input_shape, self.patch_output_shape, True
                )
                self_idx = np.full(patch_info.shape[0], file_idx, dtype=np.int32)
                patch_info = np.concatenate([patch_info, self_idx[:, None]], axis=-1)
                # ? may be expensive op
                patch_info = np.split(patch_info, patch_info.shape[0], axis=0)
                patch_info = [np.squeeze(p) for p in patch_info]

                # * this factor=5 is only applicable for HoVerNet
                expected_usage = sys.getsizeof(img) * 5
                available_ram -= expected_usage
                if available_ram < 0:
                    break

                file_idx += 1
                # if file_idx == 4: break
                use_path_list.append(file_path)
                cache_image_list.append(img)
                cache_patch_info_list.extend(patch_info)
                # TODO: refactor to explicit protocol
                cache_image_info_list.append([src_shape, len(patch_info), top_corner])

            # * apply neural net on cached data
            dataset = SerializeFileList(
                cache_image_list, cache_patch_info_list, self.patch_input_shape
            )
            dataloader = data.DataLoader(
                dataset,
                num_workers=self.nr_inference_workers,
                batch_size=self.batch_size,
                drop_last=False,
            )

            pbar = tqdm.tqdm(
                desc="Process Patches",
                leave=True,
                total=int(len(cache_patch_info_list) / self.batch_size) + 1,
                ncols=80,
                ascii=True,
                position=0,
            )

            accumulated_patch_output = []
            for batch_idx, batch_data in enumerate(dataloader):
                sample_data_list, sample_info_list = batch_data
                sample_output_list = self.run_step(sample_data_list)
                sample_info_list = sample_info_list.numpy()
                curr_batch_size = sample_output_list.shape[0]
                sample_output_list = np.split(
                    sample_output_list, curr_batch_size, axis=0
                )
                sample_info_list = np.split(sample_info_list, curr_batch_size, axis=0)
                sample_output_list = list(zip(sample_info_list, sample_output_list))
                accumulated_patch_output.extend(sample_output_list)
                pbar.update()
            pbar.close()

            # * parallely assemble the processed cache data for each file if possible
            future_list = []
            for file_idx, file_path in enumerate(use_path_list):
                image_info = cache_image_info_list[file_idx]
                file_ouput_data, accumulated_patch_output = detach_items_of_uid(
                    accumulated_patch_output, file_idx, image_info[1]
                )

                # * detach this into func and multiproc dispatch it
                src_pos = image_info[2]  # src top left corner within padded image
                src_image = cache_image_list[file_idx]
                src_image = src_image[
                    src_pos[0] : src_pos[0] + image_info[0][0],
                    src_pos[1] : src_pos[1] + image_info[0][1],
                ]

                base_name = pathlib.Path(file_path).stem
                file_info = {
                    "src_shape": image_info[0],
                    "src_image": src_image,
                    "name": base_name,
                }

                post_proc_kwargs = {
                    "nr_types": self.nr_types,
                    "return_centroids": True,
                }  # dynamicalize this

                overlay_kwargs = {
                    "draw_dot": self.draw_dot,
                    "type_colour": self.type_info_dict,
                    "line_thickness": 2,
                }
                func_args = (
                    self.post_proc_func,
                    post_proc_kwargs,
                    file_ouput_data,
                    file_info,
                    overlay_kwargs,
                )

                # dispatch for parallel post-processing
                if proc_pool is not None:
                    proc_future = proc_pool.submit(_post_process_patches, *func_args)
                    # ! manually poll future and call callback later as there is no guarantee
                    # ! that the callback is called from main thread
                    future_list.append(proc_future)
                else:
                    proc_output = _post_process_patches(*func_args)
                    proc_callback(proc_output)

            if proc_pool is not None:
                # loop over all to check state a.k.a polling
                for future in as_completed(future_list):
                    # TODO: way to retrieve which file crashed ?
                    # ! silent crash, cancel all and raise error
                    if future.exception() is not None:
                        log_info("Silent Crash")
                        # ! cancel somehow leads to cascade error later
                        # ! so just poll it then crash once all future
                        # ! acquired for now
                        # for future in future_list:
                        #     future.cancel()
                        # break
                    else:
                        file_path = proc_callback(future.result())
                        log_info("Done Assembling %s" % file_path)
        return




"""
Converts from HoVernet json to QuPath geojson.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Dict, Any, Tuple, List, Optional
import argparse
import geojson
from tqdm import tqdm
from ..utils.preprocessing import parse_path, get_names, create_dir, read_json, create_geojson


Point = Tuple[float, float]
Contour = List[Point]


def parse_contours(nuc: Dict[str, Any], num_classes: Optional[int] = 2) -> List[Contour]:
    """
    Parses contours of cells from the given HoverNet JSON dictionary.

    :param nuc: A dictionary containing HoverNet nuclei information.
    :type nuc: Dict[str, Any]
    :param num_classes: The number of classes (default 2).
    :type num_classes: Optional[int]
    :return: A list of contours of cells as list of points. Each contour has the same point at position 0 and -1.
    :rtype: List[Contour]

    Each contour has the same point at position 0 and -1.
    If a cell has no class information, it is assumed to belong to the "segmented" class (class 3).
    """
    contours_ = []
    for inst in nuc:
        inst_info = nuc[inst]
        inst_type = inst_info['type']
        if inst_type >= 1 and inst_type <= num_classes:
            inst_contour = inst_info['contour']
            inst_contour.append(inst_contour[0])
            contours_.append((inst_contour, inst_type))
        elif inst_type is None:
            inst_contour = inst_info['contour']
            inst_contour.append(inst_contour[0])
            contours_.append((inst_contour, 3))
    return contours_


def save_contours(out_dir: str, name: str, contours: List[Contour]) -> None:
    """
    Save geojson in a file with given name in out_dir.

    :param out_dir: The directory where the file will be saved.
    :type out_dir: str
    :param name: The name of the file.
    :type name: str
    :param contours: The list of contours in the GeoJSON format.
    :type contours: List[Contour]
    """
    create_dir(parse_path(out_dir))
    with open(out_dir + name + '.geojson', 'w') as f:
        geojson.dump(contours, f, sort_keys=True, indent=2)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--json-dir', type=str, default='./',
                        help='Path to json files.')
    parser.add_argument('--gson-dir', type=str, default='./',
                        help='Path to save files.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main_with_args(args):
    json_dir = parse_path(args.json_dir)
    names = get_names(json_dir, '.json')
    for name in tqdm(names):
        json_path = json_dir + name + '.json'
        nuc = read_json(json_path)
        contours = parse_contours(nuc, args.num_classes)
        features = create_geojson(contours, args.num_classes)
        save_contours(parse_path(args.gson_dir), name, features)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



import numpy as np
from scipy.ndimage import measurements
from skimage import morphology as morph
import matplotlib.pyplot as plt

from ...misc.utils import center_pad_to_shape, cropping_center, get_bounding_box
from ...dataloader.augs import fix_mirror_padding


####
def gen_instance_hv_map(ann, crop_shape):
    """Input annotation must be of original shape.
    
    The map is calculated only for instances within the crop portion
    but based on the original shape in original image.

    Perform following operation:
    Obtain the horizontal and vertical distance maps for each
    nuclear instance.

    """
    orig_ann = ann.copy()  # instance ID map
    fixed_ann = fix_mirror_padding(orig_ann)
    # re-cropping with fixed instance id map
    crop_ann = cropping_center(fixed_ann, crop_shape)
    # TODO: deal with 1 label warning
    crop_ann = morph.remove_small_objects(crop_ann, min_size=30)

    x_map = np.zeros(orig_ann.shape[:2], dtype=np.float32)
    y_map = np.zeros(orig_ann.shape[:2], dtype=np.float32)

    inst_list = list(np.unique(crop_ann))
    if 0 in inst_list:
        inst_list.remove(0)  # 0 is background
    for inst_id in inst_list:
        inst_map = np.array(fixed_ann == inst_id, np.uint8)
        inst_box = get_bounding_box(inst_map)

        # expand the box by 2px
        # Because we first pad the ann at line 207, the bboxes
        # will remain valid after expansion
        inst_box[0] -= 2
        inst_box[2] -= 2
        inst_box[1] += 2
        inst_box[3] += 2

        inst_map = inst_map[inst_box[0] : inst_box[1], inst_box[2] : inst_box[3]]

        if inst_map.shape[0] < 2 or inst_map.shape[1] < 2:
            continue

        # instance center of mass, rounded to nearest pixel
        inst_com = list(measurements.center_of_mass(inst_map))

        try:
            inst_com[0] = int(inst_com[0] + 0.5)
            inst_com[1] = int(inst_com[1] + 0.5)
        except:
            continue

        inst_x_range = np.arange(1, inst_map.shape[1] + 1)
        inst_y_range = np.arange(1, inst_map.shape[0] + 1)
        # shifting center of pixels grid to instance center of mass
        inst_x_range -= inst_com[1]
        inst_y_range -= inst_com[0]

        inst_x, inst_y = np.meshgrid(inst_x_range, inst_y_range)

        # remove coord outside of instance
        inst_x[inst_map == 0] = 0
        inst_y[inst_map == 0] = 0
        inst_x = inst_x.astype("float32")
        inst_y = inst_y.astype("float32")

        # normalize min into -1 scale
        if np.min(inst_x) < 0:
            inst_x[inst_x < 0] /= -np.amin(inst_x[inst_x < 0])
        if np.min(inst_y) < 0:
            inst_y[inst_y < 0] /= -np.amin(inst_y[inst_y < 0])
        # normalize max into +1 scale
        if np.max(inst_x) > 0:
            inst_x[inst_x > 0] /= np.amax(inst_x[inst_x > 0])
        if np.max(inst_y) > 0:
            inst_y[inst_y > 0] /= np.amax(inst_y[inst_y > 0])

        ####
        x_map_box = x_map[inst_box[0] : inst_box[1], inst_box[2] : inst_box[3]]
        x_map_box[inst_map > 0] = inst_x[inst_map > 0]

        y_map_box = y_map[inst_box[0] : inst_box[1], inst_box[2] : inst_box[3]]
        y_map_box[inst_map > 0] = inst_y[inst_map > 0]

    hv_map = np.dstack([x_map, y_map])
    return hv_map


####
def gen_targets(ann, crop_shape, **kwargs):
    """Generate the targets for the network."""
    hv_map = gen_instance_hv_map(ann, crop_shape)
    np_map = ann.copy()
    np_map[np_map > 0] = 1

    hv_map = cropping_center(hv_map, crop_shape)
    np_map = cropping_center(np_map, crop_shape)

    target_dict = {
        "hv_map": hv_map,
        "np_map": np_map,
    }

    return target_dict


####
def prep_sample(data, is_batch=False, **kwargs):
    """
    Designed to process direct output from loader
    """
    cmap = plt.get_cmap("jet")

    def colorize(ch, vmin, vmax, shape):
        ch = np.squeeze(ch.astype("float32"))
        ch = ch / (vmax - vmin + 1.0e-16)
        # take RGB from RGBA heat map
        ch_cmap = (cmap(ch)[..., :3] * 255).astype("uint8")
        ch_cmap = center_pad_to_shape(ch_cmap, shape)
        return ch_cmap

    def prep_one_sample(data):
        shape_array = [np.array(v.shape[:2]) for v in data.values()]
        shape = np.maximum(*shape_array[:2])
        viz_list = []
        viz_list.append(colorize(data["np_map"], 0, 1, shape))
        # map to [0,2] for better visualisation.
        # Note, [-1,1] is used for training.
        viz_list.append(colorize(data["hv_map"][..., 0] + 1, 0, 2, shape))
        viz_list.append(colorize(data["hv_map"][..., 1] + 1, 0, 2, shape))
        img = center_pad_to_shape(data["img"], shape)
        return np.concatenate([img] + viz_list, axis=1)

    # cmap may randomly fails if of other types
    if is_batch:
        viz_list = []
        data_shape = list(data.values())[0].shape
        for batch_idx in range(data_shape[0]):
            sub_data = {k : v[batch_idx] for k, v in data.items()}
            viz_list.append(prep_one_sample(sub_data))
        return np.concatenate(viz_list, axis=0)
    else:
        return prep_one_sample(data)



"""
Semi-Supervised Classification with Graph Convolutional Networks, 
with some minor modifications.

References
----------
Paper: https://arxiv.org/abs/1609.02907

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""

import torch.nn as nn
import torch.nn.functional as F
from dgl.nn import GraphConv
from .norm import Norm

class GCN(nn.Module):
    def __init__(self, in_feats, h_feats, num_classes, num_layers, drop_rate, norm_type, enable_background=False):
        super(GCN, self).__init__()
        self.conv_layers = nn.ModuleList()
        self.conv_layers.append(GraphConv(in_feats, h_feats, activation=F.elu))
        self.conv_layers.append(nn.Dropout(drop_rate)) # Feature map dropout
        self.conv_layers.append(Norm(norm_type=norm_type, hidden_dim=h_feats))
        for l in range(1,num_layers):
            self.conv_layers.append(GraphConv(h_feats, h_feats, activation=F.elu))
            self.conv_layers.append(nn.Dropout(drop_rate))
            self.conv_layers.append(Norm(norm_type=norm_type, hidden_dim=h_feats))
        self.conv_layers.append(GraphConv(h_feats, num_classes))

        self.enable_background = enable_background
        if enable_background:
            self.bkgr_head = GraphConv(h_feats, 2)


    def forward(self, g, in_feat):
        h = in_feat
        for i, layer in enumerate(self.conv_layers):
            if i == len(self.conv_layers) - 1:
                if self.enable_background:  # Last layer
                    h_bkgr = self.bkgr_head(g, h)
                    h = layer(g, h)
                    return h, h_bkgr 
                h = layer(g, h)
            else:
                if i % 3 == 1:
                    h = layer(h)  # Dropout
                else:
                    h = layer(g, h)  # Other layers
        return h


import operator
import json
import torch

####
class BaseCallbacks(object):
    def __init__(self):
        self.engine_trigger = False

    def reset(self):
        pass

    def run(self, state, event):
        pass


####
class TrackLr(BaseCallbacks):
    """
    Add learning rate to tracking
    """

    def __init__(self, per_n_epoch=1, per_n_step=None):
        super().__init__()
        self.per_n_epoch = per_n_epoch
        self.per_n_step = per_n_step

    def run(self, state, event):
        # logging learning rate, decouple into another callback?
        run_info = state.run_info
        for net_name, net_info in run_info.items():
            lr = net_info["optimizer"].param_groups[0]["lr"]
            state.tracked_step_output["scalar"]["lr-%s" % net_name] = lr
        return


####
class ScheduleLr(BaseCallbacks):
    """Trigger all scheduler."""

    def __init__(self):
        super().__init__()

    def run(self, state, event):
        # logging learning rate, decouple into another callback?
        run_info = state.run_info
        for net_name, net_info in run_info.items():
            net_info["lr_scheduler"].step()
        return


####
class TriggerEngine(BaseCallbacks):
    def __init__(self, triggered_engine_name, nr_epoch=1):
        self.engine_trigger = True
        self.triggered_engine_name = triggered_engine_name
        self.triggered_engine = None
        self.nr_epoch = nr_epoch

    def run(self, state, event):
        self.triggered_engine.run(
            chained=True, nr_epoch=self.nr_epoch, shared_state=state
        )
        return


####
class PeriodicSaver(BaseCallbacks):
    """Must declare save dir first in the shared global state of the attached engine."""

    def __init__(self, per_n_epoch=1, per_n_step=None):
        super().__init__()
        self.per_n_epoch = per_n_epoch
        self.per_n_step = per_n_step

    def run(self, state, event):
        if not state.logging:
            return

        # TODO: add switch so that only one of [per_n_epoch / per_n_step] can run
        if state.curr_epoch % self.per_n_epoch != 0:
            return

        for net_name, net_info in state.run_info.items():
            net_checkpoint = {}
            for key, value in net_info.items():
                if key != "extra_info" and key != "use_cpu":
                    net_checkpoint[key] = value.state_dict()
            torch.save(
                net_checkpoint,
                "%s/%s_epoch=%d.tar" % (state.log_dir, net_name, state.curr_epoch),
            )
        return


####
class ConditionalSaver(BaseCallbacks):
    """Must declare save dir first in the shared global state of the attached engine."""

    def __init__(self, metric_name, comparator=">="):
        super().__init__()
        self.metric_name = metric_name
        self.comparator = comparator

    def run(self, state, event):
        if not state.logging:
            return

        ops = {
            ">": operator.gt,
            "<": operator.lt,
            ">=": operator.ge,
            "<=": operator.le,
        }
        op_func = ops[self.comparator]
        if self.comparator == ">" or self.comparator == ">=":
            best_value = -float("inf")
        else:
            best_value = +float("inf")

        # json stat log file, update and overwrite
        with open(state.log_info["json_file"]) as json_file:
            json_data = json.load(json_file)

        for epoch, epoch_stat in json_data.items():
            epoch_value = epoch_stat[self.metric_name]
            if op_func(epoch_value, best_value):
                best_value = epoch_value

        current_value = json_data[str(state.curr_epoch)][self.metric_name]
        if not op_func(current_value, best_value):
            return  # simply return because not satisfy

        print(
            state.curr_epoch
        )  # TODO: better way to track which optimal epoch is saved
        for net_name, net_info in state.run_info.items():
            net_checkpoint = {}
            for key, value in net_info.items():
                if key != "extra_info" and key != "use_cpu":
                    net_checkpoint[key] = value.state_dict()
            torch.save(
                net_checkpoint,
                "%s/%s_best=[%s].tar" % (state.log_dir, net_name, self.metric_name),
            )
        return


####
class AccumulateRawOutput(BaseCallbacks):
    def run(self, state, event):
        step_output = state.step_output["raw"]
        accumulated_output = state.epoch_accumulated_output

        for key, step_value in step_output.items():
            if key in accumulated_output:
                accumulated_output[key].extend(list(step_value))
            else:
                accumulated_output[key] = list(step_value)
        return


####
class ScalarMovingAverage(BaseCallbacks):
    """Calculate the running average for all scalar output of 
    each runstep of the attached RunEngine."""

    def __init__(self, alpha=0.95):
        super().__init__()
        self.alpha = alpha
        self.tracking_dict = {}

    def run(self, state, event):
        # TODO: protocol for dynamic key retrieval for EMA
        step_output = state.step_output["EMA"]

        for key, current_value in step_output.items():
            if key in self.tracking_dict:
                old_ema_value = self.tracking_dict[key]
                # calculate the exponential moving average
                new_ema_value = (
                    old_ema_value * self.alpha + (1.0 - self.alpha) * current_value
                )
                self.tracking_dict[key] = new_ema_value
            else:  # init for variable which appear for the first time
                new_ema_value = current_value
                self.tracking_dict[key] = new_ema_value

        state.tracked_step_output["scalar"] = self.tracking_dict
        return


####
class ProcessAccumulatedRawOutput(BaseCallbacks):
    def __init__(self, proc_func, per_n_epoch=1):
        # TODO: allow dynamically attach specific procesing for `type`
        super().__init__()
        self.per_n_epoch = per_n_epoch
        self.proc_func = proc_func

    def run(self, state, event):
        current_epoch = state.curr_epoch
        # if current_epoch % self.per_n_epoch != 0: return
        raw_data = state.epoch_accumulated_output
        track_dict = self.proc_func(raw_data)
        # update global shared states
        state.tracked_step_output = track_dict
        return


####
class VisualizeOutput(BaseCallbacks):
    def __init__(self, proc_func, per_n_epoch=1):
        super().__init__()
        # TODO: option to dump viz per epoch or per n step
        self.per_n_epoch = per_n_epoch
        self.proc_func = proc_func

    def run(self, state, event):
        current_epoch = state.curr_epoch
        raw_output = state.step_output["raw"]
        viz_image = self.proc_func(raw_output)
        state.tracked_step_output["image"]["output"] = viz_image
        return



"""
Script to evaluate predictions.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Dict, List, Optional
import os
import pandas as pd
import numpy as np
from sklearn.calibration import calibration_curve
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from ..utils.preprocessing import parse_path, create_dir
from ..utils.classification import metrics_from_predictions

import argparse


def abline(slope: float, intercept: float, axes: Axes) -> None:
    """
    Plot a line on the current axis given a slope and intercept.

    :param slope: The slope of the line.
    :type slope: float
    :param intercept: The y-intercept of the line.
    :type intercept: float
    :param axes: The matplotlib axes object on which to plot the line.
    :type axes: matplotlib.axes.Axes
    """
    x_vals = np.array(axes.get_xlim())
    y_vals = intercept + slope * x_vals
    axes.plot(x_vals, y_vals, linestyle='dotted', color='k')


def draw_reliability_diagram(
        y_true: np.ndarray,
        y_probs: np.ndarray,
        save_path: str,
        method_name: str
        ) -> None:
    """
    Draws a reliability diagram and saves it to the specified path.

    :param y_true: The true labels.
    :type y_true: np.ndarray
    :param y_probs: The predicted probabilities.
    :type y_probs: np.ndarray
    :param save_path: The path to save the reliability diagram (without extension).
    :type save_path: str
    :param method_name: The name of the method for the legend.
    :type method_name: str
    """
    prob_true, prob_pred = calibration_curve(y_true, y_probs, n_bins=20)

    fig, ax = plt.subplots(1, 2, figsize=(16, 6))
    ax[0].plot(prob_pred, prob_true, marker='s', color='forestgreen')
    abline(1, 0, ax[0])
    ax[0].set(
        title='Reliability diagram',
        xlabel='Mean predicted probability',
        ylabel='Fraction of positives'
    )
    ax[0].legend([method_name, 'Perfectly calibrated'])

    ax[1].hist(
        y_probs,
        range=(0, 1),
        bins=20,
        label=method_name,
        color='forestgreen',
    )
    ax[1].set(title=method_name, xlabel="Mean predicted probability", ylabel="Count")

    fig.savefig(save_path + '.png')
    plt.close(fig)


def compute_metrics(
        nodes_df: pd.DataFrame,
        draw_on: Optional[str] = None,
        method_name: Optional[str] = 'Method'
        ) -> Dict[str, float]:
    """
    Computes various evaluation metrics based on the provided predictions and ground truth labels.

    This function computes the following metrics:
        - Accuracy
        - F1-score
        - ROC AUC
        - Expected Calibration Error (ECE)
        - Percentage error

    The input DataFrame must contain the following columns:
        - 'class': The ground truth labels (1 for negative, 2 for positive).
        - 'prob1': The predicted probabilities for the positive class.

    Optionally, a reliability diagram can be generated and saved to a file specified by 'draw_on'.
    The 'method_name' parameter is used for the legend in the reliability diagram.

    :param nodes_df: The DataFrame containing the predictions and ground truth labels.
    :type nodes_df: pd.DataFrame
    :param draw_on: The file path to save the reliability diagram (without extension), defaults to None.
    :type draw_on: Optional[str]
    :param method_name: The name of the method for the reliability diagram legend, defaults to 'Method'.
    :type method_name: Optional[str]

    :return: A dictionary containing the computed metrics.
    :rtype: Dict[str, float]
    """
    labels = nodes_df['class'].to_numpy() - 1
    probs = nodes_df['prob1'].to_numpy()
    preds = (probs > 0.5) * 1

    acc, f1, auc, perc_error, ece = metrics_from_predictions(labels, preds, probs, 2)

    if draw_on is not None:
        draw_reliability_diagram(labels, probs, draw_on, method_name)

    return {'Accuracy': acc, 'F1-score': f1, 'ROC AUC': auc, 'Error percentage': perc_error, 'ECE': ece}


def save_metrics(metrics: Dict[str, float], save_name):
    """
    Saves the computed metrics into a CSV file.

    :param metrics: A dictionary containing the computed metrics.
    :type metrics: Dict[str, float]
    :param save_name: The file path to save the metrics (without extension).
    :type save_name: str
    """
    metrics_df = pd.DataFrame(metrics)
    metrics_df.to_csv(save_name + '.csv', index=False)


def join_dictionaries(dict1: Dict[str, List[float]], dict2: Dict[str, float]) -> None:
    """
    Joins the values from dict2 into dict1, modifying dict1.

    For each key in dict2, the corresponding value is added to the list in dict1
    with the same key. If dict1 does not have the key, it is created with an empty list.

    :param dict1: The dictionary to be modified.
    :type dict1: Dict[str, List[float]]
    :param dict2: The dictionary containing the values to be joined into dict1.
    :type dict2: Dict[str, float]
    """
    for key in dict2.keys():
        if key not in dict1:
            dict1[key] = []
        dict1[key].append(dict2[key])


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--node-dir', type=str, required=True,
                        help='Folder with .nodes.csv prediction files.')
    parser.add_argument('--save-file', type=str, required=True,
                        help='Name to file where to save the results. Must not contain extension.')
    parser.add_argument('--by-img', action='store_true',
                        help='Whether to evaluate images separately or totally.')
    parser.add_argument('--draw', action='store_true',
                        help='Whether to draw reliability diagrams.')
    parser.add_argument('--draw-dir', type=str,
                        help='Folder to save reliability diagram on by-img mode.')
    parser.add_argument('--name', type=str, default='Method',
                        help='The name of the model used. Default: Method.')
    return parser


def main_with_args(args):
    node_files = os.listdir(args.node_dir)
    node_paths = [parse_path(args.node_dir) + x for x in node_files]
    metrics = {}
    if args.by_img:
        for node_path, node_file in zip(node_paths, node_files):
            nodes_df = pd.read_csv(node_path)
            if args.draw:
                assert args.draw_dir is not None, 'In by-img mode you must provide a folder for saving drawings.'
                draw_dir = parse_path(args.draw_dir)
                create_dir(draw_dir)
                aux_metrics = compute_metrics(nodes_df, draw_on=draw_dir + node_file, method_name=args.name)
            else:
                aux_metrics = compute_metrics(nodes_df)
            aux_metrics['name'] = node_file[:-10]
            join_dictionaries(metrics, aux_metrics)
    else:
        total_nodes_df = pd.DataFrame()
        for node_path in node_paths:
            nodes_df = pd.read_csv(node_path)
            total_nodes_df = pd.concat([total_nodes_df, nodes_df])
        total_nodes_df.reset_index(inplace=True)
        total_nodes_df.drop('index', axis=1, inplace=True)
        if args.draw:
            aux_metrics = compute_metrics(total_nodes_df, draw_on=args.save_file, method_name=args.name)
        else:
            aux_metrics = compute_metrics(total_nodes_df)
        join_dictionaries(metrics, aux_metrics)
    save_metrics(metrics, args.save_file)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os

from tumourkit.utils.preprocessing import parse_path, get_names, read_json

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
JSON_DIR = parse_path(TEST_DIR) + 'json/'

@pytest.mark.parametrize('name', get_names(JSON_DIR, '.json'))
def test_hov_prob(name):
    nuc = read_json(os.path.join(JSON_DIR, name + '.json'))
    for inst in nuc:
        inst_info = nuc[inst]
        inst_centroid = inst_info['centroid']
        inst_type = inst_info['type']
        if inst_type != 0:
            prob1 = inst_info['prob1']
            pred_type = (prob1 > 0.5) * 1 + 1
            assert int(inst_type) == int(pred_type)


"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import cProfile
import io
import pstats
import os
import json
from postprocessing.rswoosh import rswoosh
from ..utils.preprocessing import get_names, parse_path
from ..utils.postprocessing import create_comparator, merge_cells


PROF_DIR = os.path.dirname(os.path.abspath(__file__))
PKG_DIR = os.path.dirname(PROF_DIR)
RSWOOSH_DIR = parse_path(PKG_DIR) + 'tests/rswoosh/'
THRESHOLD = 1
NUM_FRONTIER = 15
names = get_names(RSWOOSH_DIR, '.input.json')


def exec_conversion(name):
    with open(RSWOOSH_DIR + name + '.input.json', 'r') as f:
        inp = json.load(f)
    with open(RSWOOSH_DIR + name + '.output.json', 'r') as f:
        expected_out = json.load(f)
    predicted_out = rswoosh(inp, create_comparator(THRESHOLD, NUM_FRONTIER), merge_cells)
    if not len(predicted_out) == len(expected_out):
        print(name)
        return False
    return True


for name in names:
    pr = cProfile.Profile()
    pr.enable()

    my_result = exec_conversion(name)

    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('tottime')
    ps.print_stats()

    with open('rswoosh/rswoosh_profile_' + name + '.txt', 'w+') as f:
        f.write(s.getvalue())



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import pandas as pd
import numpy as np

from tumourkit.preprocessing.graph2centroids import graph2centroids
from tumourkit.utils.preprocessing import get_names, parse_path, read_graph, read_centroids
from tumourkit.utils.nearest import generate_tree, find_nearest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
GRAPH_DIR = parse_path(TEST_DIR) + 'graphs/'
CENTROIDS_DIR = parse_path(TEST_DIR) + 'pngcsv/'
THRESHOLD = 0.1


def check_equal(a: pd.DataFrame, b: pd.DataFrame) -> bool:
    a_tree = generate_tree(a)
    for point in b:
        closest_idx = find_nearest(point, a_tree)
        closest = a[closest_idx]
        if closest[2] != point[2]:
            return False
    return True


@pytest.mark.parametrize("name", get_names(GRAPH_DIR, '.nodes.csv'))
def test_graph2centroids(name):
    graph_file = read_graph(name, GRAPH_DIR)
    centroids_file = graph2centroids(graph_file, num_classes=2)
    ref_centroids_file = read_centroids(name, CENTROIDS_DIR)
    if not check_equal(centroids_file, ref_centroids_file):
        assert False, name
    assert(True)



import tqdm
from enum import Enum


####
class Events(Enum):
    EPOCH_STARTED = "epoch_started"
    EPOCH_COMPLETED = "epoch_completed"
    STEP_STARTED = "step_started"
    STEP_COMPLETED = "step_completed"
    STARTED = "started"
    COMPLETED = "completed"
    EXCEPTION_RAISED = "exception_raised"


####
class State(object):
    """An object that is used to pass internal and user-defined state between event handlers."""

    def __init__(self):
        # settings propagated from config
        self.logging = None
        self.log_dir = None
        self.log_info = None

        # internal variable
        self.curr_epoch_step = 0  # current step in epoch
        self.curr_global_step = 0  # current global step
        self.curr_epoch = 0  # current global epoch

        # TODO: [LOW] better document this
        # for outputing value that will be tracked per step
        # "scalar" will always be printed out and added to the tensorboard
        # "images" will need dedicated function to process and added to the tensorboard

        # ! naming should match with types supported for serialize
        # TODO: Need way to dynamically adding new types
        self.tracked_step_output = {
            "scalar": {},  # type : {variable_name : variablee_value}
            "image": {},
        }
        # TODO: find way to known which method bind/interact with which value

        self.epoch_accumulated_output = {}  # all output of the current epoch

        # TODO: soft reset for pertain variable for N epochs
        self.run_accumulated_output = []  # of run until reseted

        # holder for output returned after current runstep
        # * depend on the type of training i.e GAN, the updated accumulated may be different
        self.step_output = None

        self.global_state = None
        return

    def reset_variable(self):
        # type : {variable_name : variable_value}
        self.tracked_step_output = {k: {} for k in self.tracked_step_output.keys()}

        # TODO: [CRITICAL] refactor this
        if self.curr_epoch % self.pertain_n_epoch_output == 0:
            self.run_accumulated_output = []

        self.epoch_accumulated_output = {}

        # * depend on the type of training i.e GAN, the updated accumulated may be different
        self.step_output = None  # holder for output returned after current runstep
        return


####
class RunEngine(object):
    """
    TODO: Include docstring
    """

    def __init__(
        self,
        engine_name=None,
        dataloader=None,
        run_step=None,
        run_info=None,
        log_info=None,  # TODO: refactor this with trainer.py
    ):

        # * auto set all input as object variables
        self.engine_name = engine_name
        self.run_step = run_step
        self.dataloader = dataloader

        # * global variable/object holder shared between all event handler
        self.state = State()
        # * check if correctly referenced, not new copies
        self.state.attached_engine_name = engine_name  # TODO: redundant?
        self.state.run_info = run_info
        self.state.log_info = log_info
        self.state.batch_size = dataloader.batch_size

        # TODO: [CRITICAL] match all the mechanism outline with opt
        self.state.pertain_n_epoch_output = 1 if engine_name == "valid" else 1

        self.event_handler_dict = {event: [] for event in Events}

        # TODO: think about this more
        # to share global state across a chain of RunEngine such as
        # from the engine for training to engine for validation

        #
        self.terminate = False
        return

    def __reset_state(self):
        # TODO: think about this more, looks too redundant
        new_state = State()
        new_state.attached_engine_name = self.state.attached_engine_name
        new_state.run_info = self.state.run_info
        new_state.log_info = self.state.log_info
        self.state = new_state
        return

    def __trigger_events(self, event):
        for callback in self.event_handler_dict[event]:
            callback.run(self.state, event)
            # TODO: exception and throwing error with name or sthg to allow trace back
        return

    # TODO: variable to indicate output dependency between handler !
    def add_event_handler(self, event_name, handler):
        self.event_handler_dict[event_name].append(handler)

    # ! Put into trainer.py ?
    def run(self, nr_epoch=1, shared_state=None, chained=False):

        # TODO: refactor this
        if chained:
            self.state.curr_epoch = 0
        self.state.global_state = shared_state

        while self.state.curr_epoch < nr_epoch:
            try:
                self.state.reset_variable()  # * reset all EMA holder per epoch

                if not chained:
                    print("----------------EPOCH %d" % (self.state.curr_epoch + 1))

                self.__trigger_events(Events.EPOCH_STARTED)

                pbar_format = (
                    "Processing: |{bar}| "
                    "{n_fmt}/{total_fmt}[{elapsed}<{remaining},{rate_fmt}]"
                )
                if self.engine_name == "train":
                    pbar_format += (
                        "Batch = {postfix[1][Batch]:0.5f}|" "EMA = {postfix[1][EMA]:0.5f}"
                    )
                    # * changing print char may break the bar so avoid it
                    pbar = tqdm.tqdm(
                        total=len(self.dataloader),
                        leave=True,
                        initial=0,
                        bar_format=pbar_format,
                        ascii=True,
                        postfix=["", dict(Batch=float("NaN"), EMA=float("NaN"))],
                    )
                else:
                    pbar = tqdm.tqdm(
                        total=len(self.dataloader),
                        leave=True,
                        bar_format=pbar_format,
                        ascii=True,
                    )

                for data_batch in self.dataloader:
                    self.__trigger_events(Events.STEP_STARTED)

                    step_run_info = [
                        self.state.run_info,
                        {
                            "epoch": self.state.curr_epoch,
                            "step": self.state.curr_global_step,
                        },
                    ]
                    step_output = self.run_step(data_batch, step_run_info)
                    self.state.step_output = step_output

                    self.__trigger_events(Events.STEP_COMPLETED)
                    self.state.curr_global_step += 1
                    self.state.curr_epoch_step += 1

                    if self.engine_name == "train":
                        pbar.postfix[1]["Batch"] = step_output["EMA"]["overall_loss"]
                        pbar.postfix[1]["EMA"] = self.state.tracked_step_output["scalar"][
                            "overall_loss"
                        ]
                    pbar.update()
                pbar.close()  # to flush out the bar before doing end of epoch reporting
                self.state.curr_epoch += 1
                self.__trigger_events(Events.EPOCH_COMPLETED)

                # TODO: [CRITICAL] align the protocol
                self.state.run_accumulated_output.append(
                    self.state.epoch_accumulated_output
                )
            except Exception as e:
                print(e)

        return




"""
Script to generate txt files with the names of the files in the
train, validation, test split folders.

Input
-----
The path to the folders.

Output
------
Txt files with one name per line, the names don't contain the
(x,y).npy at the end.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import List
import argparse
import os
import re
from .preprocessing import parse_path, create_dir

parser = argparse.ArgumentParser()
parser.add_argument('--train', type=str,
                    help='Path to train folder as created by my_extract_patches.py')
parser.add_argument('--validation', type=str,
                    help='Path to validation folder as created by my_extract_patches.py')
parser.add_argument('--test', type=str,
                    help='Path to test folder as created by my_extract_patches.py')
parser.add_argument('--out-dir', type=str,
                    help='Path to save the txt files.')


def remove_coordinates(name: str) -> str:
    """
    Removes the (number, number) pattern and '.npy' extension from the input string.

    :param name: The input string with a (number, number) pattern and '.npy' extension.
    :type name: str
    :return: The input string with the (number, number) pattern and '.npy' extension removed.
    :rtype: str
    """
    regex = r'\([0-9]{1,3},[0-9]{1,3}\)'
    parenthesis = re.findall(regex, name)
    par_len = len(parenthesis[0])
    npy_len = len('.npy')
    return name[:-(par_len + npy_len)]


def read_files(path: str) -> List[str]:
    """
    Returns a list of file names without the (number, number).npy pattern at the end.

    :param path: The path to the directory containing the files.
    :type path: str
    :return: A list of file names without the (number, number).npy pattern at the end.
    :rtype: List[str]
    """
    files = os.listdir(path)
    file_list = []
    for file_name in files:
        file_list.append(remove_coordinates(file_name))
    file_list = list(set(file_list))
    return file_list


def save_txt(file_list: List[str], path: str, name: str) -> None:
    """
    Saves a list of file names to a text file with one name per line.

    :param file_list: The list of file names to save.
    :type file_list: list[str]
    :param path: The path to the directory where the text file will be saved.
    :type path: str
    :param name: The name of the text file to be saved.
    :type name: str
    :return: None
    """
    with open(path + name, 'a') as f:
        for file_name in file_list:
            print(file_name, file=f)


if __name__ == '__main__':
    args = parser.parse_args()
    train_files = read_files(args.train)
    val_files = read_files(args.validation)
    test_files = read_files(args.test)

    OUT_DIR = parse_path(args.out_dir)
    create_dir(OUT_DIR)

    save_txt(train_files, OUT_DIR, 'train.txt')
    save_txt(val_files, OUT_DIR, 'validation.txt')
    save_txt(test_files, OUT_DIR, 'test.txt')



"""
Merge broken cells using a morphological algorithm
to detect touching frontiers.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import os
from typing import Callable, Dict, Tuple
import numpy as np
import pandas as pd
import skimage
from ..utils.preprocessing import (
    parse_path, create_dir, get_names, save_pngcsv, read_labels
)
import argparse
from tqdm import tqdm


MAX_CELLS = 1500


def create_id_map() -> Tuple[Callable[[np.ndarray], np.ndarray], Dict[int, Tuple[int, int]]]:
    """
    Creates an index mapping function and its inverse mapping based on a unique difference per pair.
    The index mapping function `f` is defined as `f(x) = x^5`, ensuring that the difference between function values for each pair of indices is unique.
    Extra information: https://math.stackexchange.com/questions/4565014/injectivity-of-given-integer-function/4567383#4567383

    :return: A tuple containing the index mapping function `f` and the inverse mapping of the differences.
    :rtype: Tuple[Callable[[np.ndarray], np.ndarray], Dict[int, Tuple[int, int]]]
    """
    def f(x: int) -> int:
        return x**5
    sq_dif = []
    inv_diff_mapping = {}
    for i in range(MAX_CELLS):
        for j in range(i + 1, MAX_CELLS):
            sq_dif.append(f(j) - f(i))
            inv_diff_mapping[f(j) - f(i)] = [j, i]
    assert (len(sq_dif) - len(set(sq_dif))) == 0, 'Defined function is not injective.'

    vec_mapping = np.vectorize(f)
    return vec_mapping, inv_diff_mapping


def get_gradient(png: np.ndarray) -> np.ndarray:
    """"
    Computes the morphological gradient of a binary image.

    :param png: The binary image for which the gradient is computed.
    :type png: np.ndarray

    :return: The gradient image.
    :rtype: np.ndarray
    """
    mask = png.copy()
    mask[mask > 0] = 1
    dilation = skimage.morphology.dilation(png, np.ones((3, 3)))
    grad = dilation - png
    grad_bkgr = grad * mask
    return grad_bkgr


def merge_cells(
        png: np.ndarray,
        vec_mapping: Callable[[np.ndarray], np.ndarray],
        inv_diff_mapping: Dict[int, Tuple[int, int]]
        ) -> np.ndarray:
    """
    Merges cells in a binary image based on the shared frontier length.

    :param png: The binary image containing cells.
    :type png: np.ndarray
    :param vec_mapping: The vectorized mapping function to map indices.
    :type vec_mapping: Callable[[np.ndarray], np.ndarray]
    :param inv_diff_mapping: The inverse mapping of differences between indices.
    :type inv_diff_mapping: Dict[int, Tuple[int, int]]

    :return: The merged binary image.
    :rtype: np.ndarray
    """
    png_cp = png.copy()
    png_cp = vec_mapping(png_cp)  # Indices mapping
    grad_bkgr = get_gradient(png_cp)  # Morphological gradient
    # Select pairs with long frontier
    unique, counts = np.unique(grad_bkgr, return_counts=True)
    component = [-1 for _ in range(MAX_CELLS)]
    for el, freq in zip(unique, counts):
        if el > 0 and freq > 13:  # More than 13 points in frontier
            pair = inv_diff_mapping[el]
            # Get component of each index
            if component[pair[0]] != -1:
                pair[0] = component[pair[0]]
            if component[pair[1]] != -1:
                pair[1] = component[pair[1]]
            M = min(*pair)  # Merge operation is to set both indices to the minimum
            # Update component of each index
            for k, el in enumerate(component):
                if el == pair[0] or el == pair[1]:
                    component[k] = M
            component[pair[0]] = M
            component[pair[1]] = M
            # Map indices to new component
            png[png == pair[0]] = M
            png[png == pair[1]] = M
    return png


def remove_lost_ids(png: np.ndarray, csv: pd.DataFrame) -> pd.DataFrame:
    """
    Removes identifiers in csv that are not present in the png.

    :param png: The binary image containing cell identifiers.
    :type png: np.ndarray
    :param csv: The DataFrame containing cell information with identifiers.
    :type csv: pd.DataFrame

    :return: The updated DataFrame with removed identifiers.
    :rtype: pd.DataFrame
    """
    next_ids = set(np.unique(png))
    curr_ids = set(csv.id)
    assert next_ids.difference(curr_ids) == set([0]), "Indices doesn't match"
    remove_ids = curr_ids.difference(next_ids)
    csv.set_index('id', inplace=True)
    csv.drop(remove_ids, axis='index', inplace=True)
    csv.reset_index(inplace=True)
    return csv


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--png-dir', type=str, required=True,
                        help='Path to png files.')
    parser.add_argument('--csv-dir', type=str, required=True,
                        help='Path to csv files.')
    parser.add_argument('--output-path', type=str, required=True,
                        help='Path to save files.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)
    output_path = parse_path(args.output_path)
    png_out_path = os.path.join(output_path, 'postPNG')
    csv_out_path = os.path.join(output_path, 'postCSV')
    create_dir(output_path)
    create_dir(png_out_path)
    create_dir(csv_out_path)

    names = get_names(png_dir, '.GT_cells.png')
    vec_mapping, inv_diff_mapping = create_id_map()
    for k, name in tqdm(enumerate(names)):
        png, csv = read_labels(name, png_dir, csv_dir)
        if png is None or csv is None:
            continue
        assert (np.max(csv.id) < MAX_CELLS), "Exceeded maximum number of cells."
        png = merge_cells(png, vec_mapping, inv_diff_mapping)
        csv = remove_lost_ids(png, csv)
        save_pngcsv(png, csv, png_out_path, csv_out_path, name)



"""
Pipeline to ease experiment replication.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import shutil
import argparse
from argparse import Namespace
import logging
from logging import Logger
import os
from .segmentation import hov_train, hov_infer
from .utils.pipes import check_training, WrongConfigurationError, HovernetNotFoundError, check_void
from .preprocessing import hovernet2centroids_main, geojson2pngcsv_main, pngcsv2centroids_main
from .segmentation import pngcsv2npy
from .utils.preprocessing import get_names
from . import eval_segment
from .classification import train_xgb, train_gnn


def run_preprocessing(args: Namespace, logger: Logger) -> None:
    """
    Runs the preprocessing steps.

    :param args: The arguments for running the preprocessing.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    if args.experiment == 'cnn-gnn' or args.experiment == 'xgb-gnn' or args.experiment == 'void-gnn':
        check_training(args, logger)
    os.makedirs(args.output_dir, exist_ok=True)
    if args.experiment == 'scaling':
        if args.pretrained_path is None:
            raise WrongConfigurationError('You must provide a path to pretrained Hovernet weights for the scaling experiment.')
    return


def hovernet_preproc_with_shape(shape: str, args: Namespace, logger: Logger) -> None:
    """
    Converts the Hovernet gson format to the rest of formats.

    :param shape: The shape of the Hovernet model.
    :type shape: str

    :param args: The arguments for running Hovernet preprocessing.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    real_shape = shape[:-2] if 'FT' in shape else shape
    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'data', real_shape), exist_ok=True)
    for split in ['train', 'validation', 'test']:
        logger.info(f'Parsing {split} split')
        if check_void(os.path.join(args.root_dir, 'data', split, 'png')) or check_void(os.path.join(args.root_dir, 'data', split, 'csv')):
            logger.info('   From geojson to pngcsv.')
            newargs = Namespace(
                gson_dir=os.path.join(args.root_dir, 'data', split, 'gson'),
                png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
                num_classes=args.num_classes,
            )
            geojson2pngcsv_main(newargs)
        os.makedirs(os.path.join(args.output_dir, 'hovernet', 'data', real_shape, split), exist_ok=True)
        os.makedirs(os.path.join(args.output_dir, 'hovernet', 'data', real_shape, split, 'npy'), exist_ok=True)
        if check_void(os.path.join(args.output_dir, 'hovernet', 'data', real_shape, split, 'npy')):
            newargs = Namespace(
                orig_dir=os.path.join(args.root_dir, 'data', 'orig'),
                png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
                out_dir=os.path.join(args.output_dir, 'hovernet', 'data', real_shape, split, 'npy'),
                save_example=False, use_labels=True, split=False,
                shape=real_shape
            )
            pngcsv2npy(newargs)
    return


def train_hovernet_with_shape(shape: str, args: Namespace, logger: Logger) -> None:
    """
    Performs Hovernet preprocessing and training with the specified shape.

    :param shape: The shape of the Hovernet model.
    :type shape: str

    :param args: The arguments for Hovernet preprocessing and training.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info(f'Starting hovernet preprocessing of {shape}.')
    hovernet_preproc_with_shape(shape, args, logger)
    logger.info(f'Starting training of {shape}.')
    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'weights', shape), exist_ok=True)
    real_shape = shape[:-2] if 'FT' in shape else shape
    newargs = Namespace(
        gpu=args.gpu, view=None, save_name=None,
        log_dir=os.path.join(args.output_dir, 'hovernet', 'weights', shape),
        train_dir=os.path.join(args.output_dir, 'hovernet', 'data', real_shape, 'train', 'npy'),
        valid_dir=os.path.join(args.output_dir, 'hovernet', 'data', real_shape, 'validation', 'npy'),
        pretrained_path=args.pretrained_path if 'FT' in shape else None,
        shape=real_shape,
        num_classes=args.num_classes,
    )
    hov_train(newargs)


def infer_hovernet_with_shape(shape: str, args: Namespace, logger: Logger) -> None:
    """
    Performs inference with the specified Hovernet model shape.

    :param shape: The shape of the Hovernet model.
    :type shape: str

    :param args: The arguments for Hovernet inference.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info(f'Starting inference of {shape}.')
    newargs = {
        'nr_types': str(args.num_classes + 1),
        'type_info_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', 'type_info.json'),
        'gpu': args.gpu,
        'nr_inference_workers': '0',
        'model_path': os.path.join(args.output_dir, 'hovernet', 'weights', shape, '01', 'net_epoch=50.tar'),
        'batch_size': '10',
        'shape': shape[:-2] if 'FT' in shape else shape,
        'nr_post_proc_workers': '0',
        'model_mode': 'original',
        'help': False
    }
    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'output', shape), exist_ok=True)
    newsubargs = {
        'input_dir': os.path.join(args.root_dir, 'data', 'orig'),
        'output_dir': os.path.join(args.output_dir, 'hovernet', 'output', shape),
        'draw_dot': False,
        'save_qupath': False,
        'save_raw_map': False,
        'mem_usage': '0.2'
    }
    hov_infer(newargs, newsubargs, 'tile')


def run_postprocessing_with_shape(shape: str, args: Namespace, logger: Logger) -> None:
    """
    Performs post-processing after Hovernet training with the specified shape.

    :param shape: The shape of the Hovernet model.
    :type shape: str

    :param args: The arguments for post-processing.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Extracting Hovernet centroids from training output.')
    if not os.path.isdir(os.path.join(args.output_dir, 'hovernet', 'output', shape, 'json')):
        raise HovernetNotFoundError('Please, train again or extract hovernet outputs.')
    newargs = Namespace(
        json_dir=os.path.join(args.output_dir, 'hovernet', 'output', shape, 'json'),
        output_path=os.path.join(args.output_dir, 'hovernet', 'output', shape, 'centroids_hov'),
    )
    hovernet2centroids_main(newargs)
    for split in ['train', 'validation', 'test']:
        if check_void(os.path.join(args.root_dir, 'data', split, 'names.txt')):
            logger.info(f'Preprocessing split {split}.')
            split_names = get_names(os.path.join(args.root_dir, 'data', split, 'gson'), '.geojson')
            with open(os.path.join(args.root_dir, 'data', split, 'names.txt'), 'w') as f:
                for name in split_names:
                    print(name, file=f)
        if check_void(os.path.join(args.root_dir, 'data', split, 'png')) or check_void(os.path.join(args.root_dir, 'data', split, 'csv')):
            logger.info('   From geojson to pngcsv.')
            newargs = Namespace(
                gson_dir=os.path.join(args.root_dir, 'data', split, 'gson'),
                png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
                num_classes=args.num_classes,
            )
            geojson2pngcsv_main(newargs)
        if check_void(os.path.join(args.root_dir, 'data', split, 'centroids')):
            logger.info('   Extracting centroids from GT.')
            newargs = Namespace(
                png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
                output_path=os.path.join(args.root_dir, 'data', split, 'centroids')
            )
            pngcsv2centroids_main(newargs)


def evaluate_hovernet_with_shape(shape: str, args: Namespace, logger: Logger) -> None:
    """
    Performs evaluation of Hovernet outputs with the specified shape.

    :param shape: The shape of the Hovernet model.
    :type shape: str

    :param args: The arguments for evaluation.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info(f'Starting evaluation of {shape}.')
    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'output', shape, 'results'), exist_ok=True)
    for split in ['train', 'validation', 'test']:
        logger.info(f'    Evaluating {split} split')
        newargs = Namespace(
            names=os.path.join(args.root_dir, 'data', split, 'names.txt'),
            gt_path=os.path.join(args.root_dir, 'data', split, 'centroids'),
            pred_path=os.path.join(args.output_dir, 'hovernet', 'output', shape, 'centroids_hov'),
            save_name=os.path.join(args.output_dir, 'hovernet', 'output', shape, 'results', shape + '_' + split),
            debug_path=None,
            num_classes=args.num_classes
        )
        eval_segment(newargs, logger)


def run_scaling(args: Namespace, logger: Logger) -> None:
    """
    Runs the scaling experiment for Hovernet.

    This function performs the following steps:
        1. Trains Hovernet models with different shapes: '270', '270FT', '518', '518FT'.
        2. Performs inference with the trained models on the input data.
        3. Runs post-processing on the Hovernet output.
        4. Evaluates the Hovernet output using the ground truth data.

    :param args: The arguments for the scaling experiment.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    os.makedirs(os.path.join(args.output_dir, 'hovernet'), exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'weights'), exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'data'), exist_ok=True)
    train_hovernet_with_shape('270', args, logger)
    train_hovernet_with_shape('270FT', args, logger)
    train_hovernet_with_shape('518', args, logger)
    train_hovernet_with_shape('518FT', args, logger)

    os.makedirs(os.path.join(args.output_dir, 'hovernet', 'output'), exist_ok=True)
    infer_hovernet_with_shape('270', args, logger)
    infer_hovernet_with_shape('270FT', args, logger)
    infer_hovernet_with_shape('518', args, logger)
    infer_hovernet_with_shape('518FT', args, logger)

    run_postprocessing_with_shape('270', args, logger)
    run_postprocessing_with_shape('270FT', args, logger)
    run_postprocessing_with_shape('518', args, logger)
    run_postprocessing_with_shape('518FT', args, logger)

    evaluate_hovernet_with_shape('270', args, logger)
    evaluate_hovernet_with_shape('270FT', args, logger)
    evaluate_hovernet_with_shape('518', args, logger)
    evaluate_hovernet_with_shape('518FT', args, logger)


def run_xgb(args: Namespace, logger: Logger) -> None:
    """
    Runs the XGBoost training and evaluation.

    This function performs the following steps:
        1. Moves the graph files into a single folder for training and validation.
        2. Trains XGBoost using the merged graph files.
        3. Saves the cross-validation results.

    :param args: The arguments for the XGBoost training.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Moving graphs files into one single folder.')
    all_graphs_dir = os.path.join(args.root_dir, 'data', 'train_validation', 'graphs', 'preds')
    os.makedirs(all_graphs_dir, exist_ok=True)
    tr_graphs_dir = os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds')
    val_graphs_dir = os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds')
    files = [os.path.join(tr_graphs_dir, f) for f in os.listdir(tr_graphs_dir)] + \
            [os.path.join(val_graphs_dir, f) for f in os.listdir(val_graphs_dir)]
    for file in files:
        shutil.copy(file, all_graphs_dir)
    logger.info('Starting XGBoost training.')
    os.makedirs(os.path.join(args.output_dir, 'xgb'), exist_ok=True)
    newargs = Namespace(
        graph_dir=all_graphs_dir,
        test_graph_dir=os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        val_size=0.2,
        seed=0,
        num_workers=args.num_workers,
        cv_folds=5,
        save_name=os.path.join(args.output_dir, 'xgb', 'cv_results'),
        num_classes=args.num_classes,
    )
    train_xgb(newargs, logger)


def run_void(args: Namespace, logger: Logger) -> None:
    """
    Runs the training of GNN models without prior and morphological features.

    This function performs the following steps:
        1. Trains a GNN model without prior.
        2. Trains a GNN model without morphological features.
        3. Trains a void GNN model (without prior and morphological features).

    :param args: The arguments for the GNN training.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Training GNN without prior.')
    newargs = Namespace(
        train_node_dir=os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir=os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir=os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir=os.path.join(args.output_dir, 'gnn_no_prior_logs'),
        early_stopping_rounds=10,
        batch_size=10,
        model_name='GCN',
        save_file=os.path.join(args.output_dir, 'gnn_no_prior_logs', 'gnn_results'),
        num_confs=32,
        save_dir=os.path.join(args.root_dir, 'weights', 'classification', 'gnn_no_prior'),
        device='cpu' if args.gpu == '' else 'cuda',
        num_workers=args.num_workers,
        checkpoint_iters=-1,
        num_classes=args.num_classes,
        disable_prior=True,
        disable_morph_feats=False,
    )
    train_gnn(newargs)

    logger.info('Training GNN without morphological features.')
    newargs = Namespace(
        train_node_dir=os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir=os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir=os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir=os.path.join(args.output_dir, 'gnn_no_morph_logs'),
        early_stopping_rounds=10,
        batch_size=10,
        model_name='GCN',
        save_file=os.path.join(args.output_dir, 'gnn_no_morph_logs', 'gnn_results'),
        num_confs=32,
        save_dir=os.path.join(args.root_dir, 'weights', 'classification', 'gnn_no_morph'),
        device='cpu' if args.gpu == '' else 'cuda',
        num_workers=args.num_workers,
        checkpoint_iters=-1,
        num_classes=args.num_classes,
        disable_prior=False,
        disable_morph_feats=True,
    )
    train_gnn(newargs)

    logger.info('Training void GNN.')
    newargs = Namespace(
        train_node_dir=os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir=os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir=os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir=os.path.join(args.output_dir, 'gnn_void_logs'),
        early_stopping_rounds=10,
        batch_size=10,
        model_name='GCN',
        save_file=os.path.join(args.output_dir, 'gnn_void_logs', 'gnn_results'),
        num_confs=32,
        save_dir=os.path.join(args.root_dir, 'weights', 'classification', 'gnn_void'),
        device='cpu' if args.gpu == '' else 'cuda',
        num_workers=args.num_workers,
        checkpoint_iters=-1,
        num_classes=args.num_classes,
        disable_prior=True,
        disable_morph_feats=True,
    )
    train_gnn(newargs)


def run_cnn(args: Namespace, logger: Logger) -> None:
    raise NotImplementedError


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, help='Internal data and weights directory.', default='./.internals/')
    parser.add_argument('--output-dir', type=str, help='Folder where to save all the results.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    parser.add_argument('--experiment', type=str, choices=['scaling', 'xgb-gnn', 'void-gnn', 'cnn-gnn'])
    parser.add_argument('--pretrained-path', type=str, help='Path to initial Hovernet weights.')
    parser.add_argument('--gpu', type=str, default='0')
    parser.add_argument('--num-workers', type=int, default=0)
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('research_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    run_preprocessing(args, logger)
    if args.experiment == 'scaling':
        run_scaling(args, logger)
    if args.experiment == 'xgb-gnn':
        run_xgb(args, logger)
    if args.experiment == 'void-gnn':
        run_void(args, logger)
    if args.experiment == 'cnn-gnn':
        run_cnn(args, logger)
